import logging
from pathlib import Path

import h5py
import numpy as np

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.helpers import get_tangent_trj_str
from pysisyphus.helpers_pure import rms
from pysisyphus.intcoords.helpers import get_weighted_bond_mode
from pysisyphus.linalg import perp_comp, make_unit_vec
from pysisyphus.optimizers.closures import small_lbfgs_closure
from pysisyphus.optimizers.restrict_step import get_scale_max


class RotationConverged(Exception):
    pass


class Gaussian:
    def __init__(self, height, center, std, N):
        self.center = np.array(center)
        self.height = float(height)
        self.std = float(std)
        self.N = np.array(N)

        self.s0 = self.center.dot(self.N)

    def energy(self, R, height=None):
        if height is None:
            height = self.height

        return height * np.exp(-((R.dot(self.N) - self.s0) ** 2) / (2 * self.std**2))

    def forces(self, R, height=None):
        if height is None:
            height = self.height
        s_diff = R.dot(self.N) - self.s0

        return (
            height
            * np.exp(-(s_diff**2) / (2 * self.std**2))
            * s_diff
            / self.std**2
            * self.N
        )

    def __str__(self):
        return (
            f"Gaussian(height={self.height:.4f}, center={self.center}, "
            f"s0={self.s0:.4f}, std={self.std:.4f}"
        )


# [1] https://aip.scitation.org/doi/abs/10.1063/1.480097
#     Original Dimer paper
#     Henkelmann, 1999
# [2] https://doi.org/10.1063/1.1809574
#     Comparison of TS search methods
#     Olsen, 2004
# [3] https://doi.org/10.1063/1.2104507
#     Comparison of Dimer and P-RFO
#     Heyden, 2005
# [4] https://aip.scitation.org/doi/abs/10.1063/1.2815812
#     Superlinear Dimer method
#     Kästner, 2008


class Dimer(Calculator):
    def __init__(
        self,
        calculator,
        *args,
        N_raw=None,
        length=0.0189,
        rotation_max_cycles=15,
        rotation_method="fourier",
        rotation_thresh=1e-4,
        rotation_tol=1,
        rotation_max_element=0.001,
        rotation_interpolate=True,
        rotation_disable=False,
        rotation_disable_pos_curv=True,
        rotation_remove_trans=True,
        trans_force_f_perp=True,
        bonds=None,
        N_hessian=None,
        bias_rotation=False,
        bias_translation=False,
        bias_gaussian_dot=0.1,
        seed=None,
        write_orientations=True,
        forward_hessian=True,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.logger = logging.getLogger("dimer")

        self.calculator = calculator
        self.length = float(length)

        # Rotation parameters
        self.rotation_max_cycles = int(rotation_max_cycles)

        rotation_methods = {
            "direct": self.direct_rotation,
            "fourier": self.fourier_rotation,
        }
        try:
            self.rotation_method = rotation_methods[rotation_method]
        except KeyError as err:
            print(
                f"Invalid rotation_method={rotation_method}! Valid types are: "
                f"{tuple(self.rotation_methods.keys())}"
            )
            raise err
        self.rotation_thresh = float(rotation_thresh)
        self.rotation_tol = np.deg2rad(rotation_tol)
        self.rotation_max_element = float(rotation_max_element)
        self.rotation_interpolate = bool(rotation_interpolate)
        self.rotation_disable = bool(rotation_disable)
        self.rotation_disable_pos_curv = bool(rotation_disable_pos_curv)
        self.rotation_remove_trans = bool(rotation_remove_trans)
        self.trans_force_f_perp = trans_force_f_perp
        self.forward_hessian = forward_hessian

        # Regarding generation of initial orientation
        self.bonds = bonds
        self.N_hessian = N_hessian
        # Bias
        self.bias_rotation = bool(bias_rotation)
        self.bias_rotation_a = 0.0
        self.bias_translation = bool(bias_translation)
        self.bias_gaussian_dot = float(bias_gaussian_dot)

        self.write_orientations = write_orientations

        restrict_steps = {
            "direct": get_scale_max(self.rotation_max_element),
            "fourier": None,
        }
        self.restrict_step = restrict_steps[rotation_method]

        self._N = None
        self._coords0 = None
        self._energy0 = None
        self._f0 = None
        self._f1 = None

        self.force_evals = 0
        self.gaussians = list()

        # Set dimer direction if given
        self.N_raw = N_raw
        if self.N_raw is not None:
            msg = "Setting initial orientation from given 'N_raw'"
            if isinstance(self.N_raw, str) and Path(self.N_raw).exists():
                fn = self.N_raw
                self.N_raw = np.loadtxt(fn)
                msg = f"Read initial orientation from file '{fn}'"
                N_raw = self.N_raw.copy()
            self.N = N_raw
            self.log(msg)
        self.N_init = None

        if seed is None:
            # 2**32 - 1
            seed = np.random.randint(4294967295)
        np.random.seed(seed)
        msg = f"Using seed {seed} to initialize the random number generator.\n"
        print(msg)
        self.log(msg)

    @property
    def N(self):
        return self._N

    @N.setter
    def N(self, N_new):
        N_new = np.array(N_new, dtype=float).flatten()
        if self.rotation_remove_trans:
            N_new = self.remove_translation(N_new)
        N_new /= np.linalg.norm(N_new)
        self._N = N_new
        self._f1 = None

    @property
    def coords0(self):
        return self._coords0

    @coords0.setter
    def coords0(self, coords0_new):
        self._coords0 = coords0_new
        self._energy0 = None
        self._f0 = None
        self._f1 = None

    @property
    def coords1(self):
        return self.coords0 + self.length * self.N

    @coords1.setter
    def coords1(self, coords1_new):
        N_new = coords1_new - self.coords0
        self.N = N_new
        self._f1 = None

    @property
    def energy0(self):
        if self._energy0 is None:
            results = self.calculator.get_energy(self.atoms, self.coords0)["energy"]
            self._energy0 = results["energy"]
        return self._energy0

    @property
    def f0(self):
        if self._f0 is None:
            results = self.calculator.get_forces(self.atoms, self.coords0)
            self.force_evals += 1
            self._f0 = results["forces"]
            self._energy0 = results["energy"]
        return self._f0

    @property
    def f1(self):
        if self._f1 is None:
            results = self.calculator.get_forces(self.atoms, self.coords1)
            self.force_evals += 1
            self._f1 = results["forces"]
        return self._f1

    @f1.setter
    def f1(self, f1_new):
        self._f1 = f1_new

    @property
    def f2(self):
        """Never calculated explicitly, but estimated from f0 and f1."""
        return 2 * self.f0 - self.f1

    @property
    def can_bias_f1(self):
        return (
            self.bias_rotation
            and (self.N_init is not None)
            and (self.bias_rotation_a is not None)
            and (self.bias_rotation_a > 0.0)
        )

    @property
    def should_bias_f1(self):
        """May lead to calculation of f0 and/or f1 if present!"""
        return self.can_bias_f1 and self.C > 0.0

    @property
    def can_bias_f0(self):
        return self.bias_translation and (self.N is not None)

    @property
    def should_bias_f0(self):
        """May lead to calculation of f0 and/or f1 if present!"""
        return self.can_bias_f0 and self.C > 0.0

    @property
    def f1_bias(self):
        # Apply bias force to f1 if desired. Dont apply bias if N_init is not (yet)
        # set. When N_raw was converged to a reasonable N_init we can add
        # the bias.
        assert self.bias_rotation_a >= 0.0, "This should not be negative!"

        fN = self.bias_rotation_a * self.length * self.N.dot(self.N_init) * self.N_init

        return fN

    @property
    def rot_force(self):
        f1_perp = perp_comp(self.f1, self.N)
        f2_perp = perp_comp(self.f2, self.N)
        f_perp = f1_perp - f2_perp

        # Don't bias rotational force if curvature is already negative
        if self.should_bias_f1:
            f1_bias_perp = perp_comp(self.f1_bias, self.N)
            f_perp += f1_bias_perp

        return f_perp

    def curvature(self, f1, f2, N):
        """Curvature of the mode represented by the dimer."""
        return (f2 - f1).dot(N) / (2 * self.length)

    @property
    def C(self):
        """Shortcut for the curvature."""
        return self.curvature(self.f1, self.f2, self.N)

    def get_gaussian_energies(self, coords, sum_=True):
        energies = [gauss.energy(coords) for gauss in self.gaussians]
        if sum_:
            energies = sum(energies)
        return energies

    def get_gaussian_forces(self, coords, sum_=True):
        forces = [gauss.forces(coords) for gauss in self.gaussians]
        if sum_:
            forces = np.sum(forces, axis=0)
        return forces

    def add_gaussian(
        self, atoms, center, N, height=0.1, std=0.0529, max_cycles=50, dot_ref=None
    ):
        # Create new gaussian object with default height that will be
        # refined later.
        gaussian = Gaussian(height=height, center=center, std=std, N=N)

        if dot_ref is None:
            dot_ref = self.bias_gaussian_dot

        # Calculate real forces at inflection point of new gaussian
        infl_coords = center + gaussian.std * N
        infl_results = self.calculator.get_forces(atoms, infl_coords)
        self.force_evals += 1
        infl_forces = infl_results["forces"]

        assert (
            infl_forces.dot(N) < 0
        ), "We probably overstepped the TS. See Section 2.3 in the paper."

        forces = self.get_gaussian_forces(infl_coords) + infl_forces

        def get_dot(height):
            """Dot product of forces for a given height and orientation N."""
            tmp_forces = forces.copy()
            tmp_forces += gaussian.forces(infl_coords, height=height)
            dot = tmp_forces.dot(N)
            return dot

        def can_break(dot):
            """Convergence indicator."""
            return abs(dot - dot_ref) <= 1e-3

        def bisect(
            min_,
            max_,
        ):
            for i in range(max_cycles):
                if abs(min_ - max_) <= 1e-10:
                    raise Exception("min_ and max_ became too similar!")

                # Determie value at half of the internval
                height = min_ + (max_ - min_) / 2
                dot = get_dot(height)

                if can_break(dot):
                    break

                if dot > dot_ref:
                    max_ = height
                elif dot < dot_ref:
                    min_ = height
            return height

        # Determine appropriate height
        grow = 2
        min_height = 0
        assert get_dot(0) < dot_ref
        for i in range(max_cycles):
            dot = get_dot(height)

            if can_break(dot):
                break

            if 0 < dot < dot_ref:
                min_height = height
                height *= grow
            elif dot < dot_ref:
                height *= grow
            else:
                height = bisect(min_height, height)
                break
        dot = get_dot(height)
        gaussian.height = height
        self.gaussians.append(gaussian)

        self.log(f"Added gaussian with height={height:.6f}")
        self.log(f"There are now {len(self.gaussians)} gaussians.")

        return gaussian

    def get_N_raw_from_hessian(self, h5_fn, root=0):
        with h5py.File(h5_fn, "r") as handle:
            hessian = handle["hessian"][:]
        w, v = np.linalg.eigh(hessian)
        assert w[root] < -1e-3
        N_raw = v[:, root]
        return N_raw

    def set_N_raw(self, coords):
        self.log("No initial orientation given. Generating one.")
        if self.bonds is not None:
            N_raw = get_weighted_bond_mode(self.bonds, coords.reshape(-1, 3))
            msg = "weighted bond mode"
        elif self.N_hessian is not None:
            N_raw = self.get_N_raw_from_hessian(self.N_hessian)
            msg = "first imaginary mode of HDF5 Hessian"
        else:
            msg = "random guess"
            N_raw = np.random.rand(coords.size)
        self.log(f"Obtained initial orientation from {msg}.")
        # Make N_raw translationally invariant and normalize
        self.N = N_raw
        # Now we keep the normalized dimer orientation
        self.N_raw = self.N

    def remove_translation(self, displacement):
        # Average vector components over cartesian directions (x,y,z)
        # Sum over each direction should equal zero if translationally invariant
        average = displacement.reshape(-1, 3).mean(axis=0)

        if max(abs(average)) > 1e-8:
            self.log(
                f"N-vector not translationally invariant. Removing average before normalization."
            )
        else:
            return displacement
        # Subtract the average component along each direction to make sum zero
        invariant_displacement = (
            displacement.reshape(-1, 3) - average[None, :]
        ).flatten()
        return invariant_displacement

    def rotate_coords1(self, rad, theta):
        """Rotate dimer and produce new coords1."""
        return self.coords0 + (self.N * np.cos(rad) + theta * np.sin(rad)) * self.length

    def direct_rotation(self, optimizer, prev_step):
        rot_step = optimizer(self.rot_force, prev_step)
        rot_step = self.restrict_step(rot_step)
        # Strictly speaking rot_step should be constrained to conserve the desired
        # dimer length (coords1 - coords0)*2. This step is unconstrained.
        # Later on we calculate the actual step between the old coords1 and the new
        # coords1 that have been reconstrained.
        self.coords1 = self.coords1 + rot_step

    def fourier_rotation(self, optimizer, prev_step):
        theta_dir = optimizer(self.rot_force, prev_step)
        # Remove component that is parallel to N
        theta_dir = theta_dir - theta_dir.dot(self.N) * self.N
        theta = theta_dir / np.linalg.norm(theta_dir)

        # Get rotated endpoint geometries. The rotation takes place in a plane
        # spanned by N and theta. Theta is a unit vector perpendicular to N that
        # can be formed from the perpendicular components of the forces at the
        # endpoints.

        C = self.C
        # Derivative of the curvature, Eq. (29) in [2]
        # (f2 - f1) or -(f1 - f2)
        dC = 2 * (self.f0 - self.f1).dot(theta) / self.length
        rad_trial = -0.5 * np.arctan2(dC, 2 * abs(C))
        # self.log(f"rad_trial={rad_trial:.2f}")
        if np.abs(rad_trial) < self.rotation_tol:
            self.log(f"rad_trial={rad_trial:.2f} below threshold. Breaking.")
            raise RotationConverged

        # Trial rotation for finite difference calculation of rotational force
        # and rotational curvature.
        coords1_trial = self.rotate_coords1(rad_trial, theta)
        f1_trial = self.calculator.get_forces(self.atoms, coords1_trial)["forces"]
        self.force_evals += 1
        f2_trial = 2 * self.f0 - f1_trial
        N_trial = make_unit_vec(coords1_trial, self.coords0)
        C_trial = self.curvature(f1_trial, f2_trial, N_trial)

        b1 = 0.5 * dC
        a1 = (C - C_trial + b1 * np.sin(2 * rad_trial)) / (1 - np.cos(2 * rad_trial))
        a0 = 2 * (C - a1)

        rad_min = 0.5 * np.arctan(b1 / a1)

        # self.log(f"rad_min={rad_min:.2f}")
        def get_C(theta_rad):
            return a0 / 2 + a1 * np.cos(2 * theta_rad) + b1 * np.sin(2 * theta_rad)

        C_min = get_C(rad_min)  # lgtm [py/multiple-definition]
        if C_min > C:
            rad_min += np.deg2rad(90)
            # C_min_new = get_C(rad_min)
            # self.log("Predicted theta_min lead us to a curvature maximum "
            # f"(C(theta)={C_min:.6f}). Adding pi/2 to theta_min. "
            # f"(C(theta+pi/2)={C_min_new:.6f})"
            # )

        # TODO: handle cases where the curvature is still positive, but
        # the angle is small, so the rotation is skipped.
        # Don't do rotation for small angles
        if np.abs(rad_min) < self.rotation_tol:
            # self.log(f"rad_min={rad_min:.2f} below threshold. Breaking.")
            raise RotationConverged

        f1 = None
        # Interpolate force at coords1_rot; see Eq. (12) in [4]
        if self.rotation_interpolate:
            f1 = (
                np.sin(rad_trial - rad_min) / np.sin(rad_trial) * self.f1
                + np.sin(rad_min) / np.sin(rad_trial) * f1_trial
                + (1 - np.cos(rad_min) - np.sin(rad_min) * np.tan(rad_trial / 2))
                * self.f0
            )

        self.coords1 = self.rotate_coords1(rad_min, theta)
        self.f1 = f1

    def do_dimer_rotations(self, rotation_thresh=None):
        self.log("Doing dimer rotations")
        if rotation_thresh is None:
            rotation_thresh = self.rotation_thresh
            self.log(f"\tThreshold norm(rot_force)={rotation_thresh:.6f}")

        lbfgs = small_lbfgs_closure(gamma_mult=True)
        try:
            N_first = self.N
            prev_step = None
            for i in range(self.rotation_max_cycles):  # lgtm [py/redundant-else]
                N_cur = self.N
                rot_force = self.rot_force
                rms_rot_force = rms(rot_force)
                if self.should_bias_f1:
                    C_real = self.C
                    C_bias = -self.bias_rotation_a * (self.N.dot(self.N_init)) ** 2
                    C = C_real + C_bias
                    C_str = f"C={C: .6f}, C_real={C_real: .6f}, C_bias={C_bias: .6f}"
                else:
                    C_str = f"C={self.C: .6f}"
                self.log(f"\t{i:02d}: rms(rot_force)={rms_rot_force:.6f} {C_str}")
                if rms_rot_force <= rotation_thresh:
                    self.log("\trms(rot_force) is below threshold!")
                    raise RotationConverged
                coords1_old = self.coords1
                self.rotation_method(lbfgs, prev_step)
                actual_step = self.coords1 - coords1_old
                prev_step = actual_step
                rot_deg = np.rad2deg(np.arccos(N_cur.dot(self.N)))
                self.log(f"\t\tRotated by {rot_deg:.1f}°")
            else:
                msg = (
                    "\tDimer rotation did not converge in "
                    f"{self.rotation_max_cycles}"
                )
        except RotationConverged:
            msg = f"\tDimer rotation converged in {i+1} cycle(s)."
        self.log(msg)
        # self.log("\tN after rotation:\n\t" + str(self.N))
        self.log()
        # Restrict to interval [-1,1] where arccos is defined
        rot_deg = np.rad2deg(np.arccos(max(min(N_first.dot(self.N), 1.0), -1.0)))
        self.log(
            f"\tRotated by {rot_deg:.1f}° w.r.t. the orientation " "before rotation(s)."
        )

    def update_orientation(self, coords):
        # Generate random guess for the dimer orientation if not yet set
        if self.N is None:
            self.set_N_raw(coords)

        # Refine N_raw to N_init if not yet done
        if self.bias_rotation and self.N_init is None and self.C > 0.0:
            # Run initial sweep with a much softer convergence threshold
            self.log("Initial sweep to refine N_raw to N_init.")
            self.do_dimer_rotations(10 * self.rotation_thresh)
            if self.C > 0:
                self.N_init = self.N
                rot_rad = np.arccos(self.N_raw.dot(self.N_init))
                rot_deg = np.rad2deg(rot_rad)
                self.log(f"N_raw:\n\t{self.N_raw}")
                self.log(f"Rotated N_raw by {rot_deg:.1f}° to N_init")
                self.log(f"N_init:\n\t{self.N_init}")
                C = self.C
                self.log(f"Curvature after intial sweep is C={C:.6f}")
                self.log("Determining proper scaling factor for bias potential.")
                # Determine proper scaling factor for the quadratic bias potential
                # from the current curvature.
                scale_fact = 1.5
                self.bias_rotation_a = scale_fact * self.C
                self.log(f"Using a={scale_fact}*C={self.bias_rotation_a:.6f}")
            else:
                self.log(
                    f"Curvature after initial sweep C={self.C:.6f} is "
                    "already negative. Not setting N_init and bias_rotation_a!"
                )

        self.do_dimer_rotations()

    def get_forces(self, atoms, coords):
        self.atoms = atoms
        self.coords0 = coords

        try:
            N_backup = self.N.copy()
        except AttributeError:
            N_backup = None
        if not self.rotation_disable:
            self.update_orientation(coords)
        if (N_backup is not None) and self.rotation_disable_pos_curv and self.C > 0:
            self.log(
                "Rotation did not yield a negative curvature. "
                "Restoring previous unrotated N."
            )
            self.N = N_backup
        # Now we (have an updated self.N and) can do the force projections
        N = self.N
        # self.log(f"Orientation N:\n\t{N}")
        # Save orientation N in human-readable format, aka _trj.xyz
        # file in Angstrom.
        if self.write_orientations:
            trj_str = get_tangent_trj_str(atoms, coords, N)
            trj_fn = self.make_fn("N_trj.xyz")
            with open(trj_fn, "w") as handle:
                handle.write(trj_str)
            self.log(f"Wrote current orientation animation to '{trj_fn}'")
        # Always save orientation in Bohr
        N_fn = self.make_fn("N")
        np.savetxt(N_fn, N)

        energy = self.energy0
        self.log(f"\tenergy={self.energy0:.8f} au")

        f0 = self.f0

        if self.should_bias_f0:
            self.log("Biasing translation forces")
            self.log(f"There are currently {len(self.gaussians)} gaussians present.")
            bias_energy = self.get_gaussian_energies(coords)
            energy += bias_energy
            bias_forces = self.get_gaussian_forces(coords, sum_=False)
            try:
                bias_norms = np.linalg.norm(bias_forces, axis=1)
                bias_norms_str = np.array2string(bias_norms, precision=4)
                self.log(f"\tnorm(bias_forces)={bias_norms_str}")
            except np.AxisError:
                self.log("Skipping calculation of norm(bias_forces)")
            f0 += np.sum(bias_forces, axis=0)

        norm_f0 = np.linalg.norm(f0)
        self.log(f"\tnorm(forces)={norm_f0:.6f}")

        f_parallel = f0.dot(N) * N
        norm_parallel = np.linalg.norm(f_parallel)
        self.log(f"\tnorm(forces_parallel)={norm_parallel:.6f}")
        # self.log(f"\tforce_parallel:\n\t{f_parallel}")

        f_perp = f0 - f_parallel
        norm_perp = np.linalg.norm(f_perp)
        self.log(f"\tnorm(forces_perp)={norm_perp:.6f}")
        # self.log(f"\tforce_perp:\n\t{f_perp}")

        # Only return perpendicular component when curvature is negative
        f_tran = -f_parallel

        curv_str = "positive" if self.C > 0 else "negative"

        force_str = "reversed parallel component of"
        if (self.C < 0) or self.trans_force_f_perp:
            f_tran += f_perp
            force_str = "full"
        self.log(f"Curvature is {curv_str}. Returning {force_str} f_tran.")
        # self.log(f"\tf_tran:\n\t{f_tran}")
        self.log()

        self.calc_counter += 1

        results = {"energy": energy, "forces": f_tran}

        return results

    def get_hessian(self, atoms, coords):
        if not self.forward_hessian:
            raise Exception("Actual Hessian method not forwarded by Dimer calculator!")
        results = self.calculator.get_hessian(atoms, coords)
        return results



import scipy.constants as spc

PI = spc.pi

# Speed of light in m/s
C = spc.c
# Speed of light in atomic units
CAU = spc.value("inverse fine-structure constant")
# ħ = h / 2π in J/s
HBAR = spc.hbar
# h
PLANCK = spc.Planck
PLANCKAU = 2 * spc.pi  # With ħ_au = 1 and ħ_au=h_au/2π -> h_au = 2π
# Conversion factor for momentum in SI to atomic units
P_AU = spc.value("atomic unit of momentum")

# Bohr radius in m
BOHR2M = spc.value("Bohr radius")
# Bohr -> Å conversion factor
BOHR2ANG = BOHR2M * 1e10
# Å -> Bohr conversion factor
ANG2BOHR = 1 / BOHR2ANG
# Hartree to J
AU2J = spc.value("Hartree energy")
# Hartree to kJ / mol
AU2KJPERMOL = AU2J / 1000 * spc.Avogadro
# Hartree to kcal mol⁻¹
AU2KCALPERMOL = AU2KJPERMOL / spc.calorie
# Hartree to eV
AU2EV = spc.value("Hartree energy in eV")
# Joule to eV
JOULE2EV = AU2EV / AU2J
# Wavenumber to Hartree
NU2AU = spc.h * C * 1e2 / AU2J
# eV/Å -> Hartree/Bohr
EVANG2AUBOHR = 1 / AU2EV / ANG2BOHR
# fs -> Bohr * sqrt(amu/Hartree)
FS2AU = 0.9682885864793366
# Boltzman constant, (m² kg s⁻² K⁻¹) or just (J / K)
KB = spc.Boltzmann
KBAU = KB / AU2J
# Ideal gas constant
R = spc.gas_constant
# Atomic unit of time to second
AU2SEC = spc.value("atomic unit of time")
# Atomic mass unit to kg
AMU2KG = spc.value("atomic mass constant")
M_E = spc.electron_mass
AMU2AU = AMU2KG / M_E
# Avogadro constant
NA = spc.Avogadro

##############
# MD related #
##############

# Force/amu to acceleration
#   Hartree/(Bohr*amu) -> Bohr / fs²
FORCE2ACC = AU2J / (AMU2KG * BOHR2M**2 * 1e30)
# Velocity*amu -> Energy
#   Bohr²/fs²*amu -> Hartree
# VELO2E = AMU2KG * BOHR2M**2 / (1e-30 * AU2J)
VELO2E = 1 / FORCE2ACC
# Velocity from Bohr/fs to Bohr/t_atomic_unit
BOHRPERFS2AU = AU2SEC * 1e15



import logging

__all__ = [
    "AFIR",
    "AtomAtomTransTorque",
    "Composite",
    "ConicalIntersection",
    "DFTBp",
    "DFTD4",
    "Dimer",
    "Dummy",
    "EnergyMin",
    "EGO",
    "ExternalPotential",
    "FakeASE",
    "Gaussian09",
    "Gaussian16",
    "HardSphere",
    "IPIServer",
    "LennardJones",
    "MOPAC",
    "MultiCalc",
    "OBabel",
    "ONIOM",
    "OpenMolcas",
    "ORCA",
    "ORCA5",
    "Psi4",
    "PyPsi4",
    "PyXTB",
    "Remote",
    "TIP3P",
    "Turbomole",
    "TransTorque",
    "XTB",
    "CFOUR",
]


from pysisyphus.calculators.AFIR import AFIR
from pysisyphus.calculators.AtomAtomTransTorque import AtomAtomTransTorque
from pysisyphus.calculators.Composite import Composite
from pysisyphus.calculators.ConicalIntersection import ConicalIntersection
from pysisyphus.calculators.DFTBp import DFTBp
from pysisyphus.calculators.DFTD4 import DFTD4
from pysisyphus.calculators.Dimer import Dimer
from pysisyphus.calculators.Dummy import Dummy
from pysisyphus.calculators.EnergyMin import EnergyMin
from pysisyphus.calculators.EGO import EGO
from pysisyphus.calculators.ExternalPotential import ExternalPotential
from pysisyphus.calculators.FakeASE import FakeASE
from pysisyphus.calculators.Gaussian09 import Gaussian09
from pysisyphus.calculators.Gaussian16 import Gaussian16
from pysisyphus.calculators.IPIServer import IPIServer
from pysisyphus.calculators.HardSphere import HardSphere, PWHardSphere
from pysisyphus.calculators.LennardJones import LennardJones
from pysisyphus.calculators.MultiCalc import MultiCalc
from pysisyphus.calculators.MOPAC import MOPAC
from pysisyphus.calculators.Psi4 import Psi4
from pysisyphus.calculators.OBabel import OBabel
from pysisyphus.calculators.ONIOMv2 import ONIOM
from pysisyphus.calculators.OpenMolcas import OpenMolcas
from pysisyphus.calculators.ORCA import ORCA
from pysisyphus.calculators.ORCA5 import ORCA5
from pysisyphus.calculators.PyPsi4 import PyPsi4
from pysisyphus.calculators.PyXTB import PyXTB
from pysisyphus.calculators.Remote import Remote
from pysisyphus.calculators.TIP3P import TIP3P
from pysisyphus.calculators.TransTorque import TransTorque
from pysisyphus.calculators.Turbomole import Turbomole
from pysisyphus.calculators.XTB import XTB
from pysisyphus.calculators.CFOUR import CFOUR


logger = logging.getLogger("dimer")
logger.setLevel(logging.DEBUG)
handler = logging.FileHandler("dimer.log", mode="w", delay=True)
fmt_str = "%(message)s"
formatter = logging.Formatter(fmt_str)
handler.setFormatter(formatter)
logger.addHandler(handler)



# pdb2reaction/__init__.py

try:
    from pdb2reaction._version import __version__, __version_tuple__
except ImportError:
    __version__ = "0.0.0.dev0"
    __version_tuple__ = (0, 0, 0, "dev0")

__all__ = [
    "__version__",
    "__version_tuple__",
]



import numpy as np

from pysisyphus.optimizers.Optimizer import Optimizer


class QuickMin(Optimizer):
    def __init__(self, geometry, dt=0.35, **kwargs):
        super(QuickMin, self).__init__(geometry, **kwargs)

        self.dt = dt

    def _get_opt_restart_info(self):
        opt_restart_info = {
            "velocity": self.velocities[-1].tolist(),
        }
        return opt_restart_info

    def _set_opt_restart_info(self, opt_restart_info):
        velocity = np.array(opt_restart_info["velocity"], dtype=float)
        self.velocities = [
            velocity,
        ]

    def prepare_opt(self):
        self.velocities = [
            np.zeros_like(self.geometry.coords),
        ]

    def reset(self):
        if self.coords[-1].size != self.coords[-2].size:
            self.log(
                "Number of coordinates changed. Adapting velocity vector "
                "to new coordinate number."
            )
            self.prepare_opt()
        self.log("Resetted optimizer")

    def optimize(self):
        if self.align and self.is_cos:
            (self.velocities[-1],), _, _ = self.fit_rigid(
                vectors=(self.velocities[-1],)
            )

        prev_velocities = self.velocities[-1]
        cur_forces = self.geometry.forces
        self.forces.append(cur_forces)
        self.energies.append(self.geometry.energy)

        norm = np.linalg.norm(cur_forces)
        if not self.is_cos:
            self.log(f"Current energy={self.energies[-1]:.6f}")
        self.log(f"norm(forces)={norm:.6f}")

        if self.cur_cycle == 0:
            tmp_velocities = np.zeros_like(prev_velocities)
        else:
            overlap = prev_velocities.dot(cur_forces)
            self.log(f"Overlap of previous and current forces: {overlap:.6f}")
            if overlap > 0:
                tmp_velocities = overlap * cur_forces / cur_forces.dot(cur_forces)
            else:
                tmp_velocities = np.zeros_like(prev_velocities)
                self.log("Zeroed velocities")

        accelerations = cur_forces / self.geometry.masses_rep
        cur_velocities = tmp_velocities + self.dt * accelerations
        steps = cur_velocities * self.dt + 1 / 2 * accelerations * self.dt**2
        steps = self.scale_by_max_step(steps)
        self.velocities.append(cur_velocities)
        velo_norm = np.linalg.norm(cur_velocities)
        acc_norm = np.linalg.norm(accelerations)
        self.log(f"norm(v) = {velo_norm:.4f}, norm(a) = {acc_norm:.4f}")

        return steps



from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.intcoords.Stretch import Stretch


class DistanceFunction(Primitive):
    def __init__(self, indices, coeff=-1, **kwargs):
        self.coeff = float(coeff)
        kwargs["calc_kwargs"] = ("coeff",)
        super().__init__(indices, **kwargs)

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        return 1

    @staticmethod
    def _calculate(coords3d, indices, gradient=False, coeff=-1):
        """
        m--n  o--p
        """
        m, n, o, p = indices
        val1, grad1 = Stretch._calculate(coords3d, (m, n), gradient=True)
        val2, grad2 = Stretch._calculate(coords3d, (o, p), gradient=True)
        val = val1 + coeff * val2
        grad = grad1 + coeff * grad2
        if gradient:
            return val, grad
        return val

    @staticmethod
    def _jacobian(coords3d, indices, coeff):
        m, n, o, p = indices
        jac1 = Stretch._jacobian(coords3d, [m, n])
        jac2 = Stretch._jacobian(coords3d, [o, p])
        return jac1 + coeff * jac2



import h5py


def init_h5_group(f, group_name, data_model):
    """Create group with given name and data model."""
    group = f.create_group(group_name)
    # Create (resizable) datasets by using None in maxshape
    for key, shape in data_model.items():
        maxshape = (None,) if (len(shape) == 1) else (None, *shape[1:])
        group.create_dataset(key, shape, maxshape=maxshape)


def get_h5_group(fn, group_name, data_model=None, reset=False):
    """Return (and create if neccesary) group with given name and
    data model."""

    f = h5py.File(fn, mode="a")
    # Shortcut
    if data_model is None:
        return f[group_name]

    if group_name not in f:
        init_h5_group(f, group_name, data_model)
    group = f[group_name]

    # Check compatibility of data_model and group. If they aren't compatible
    # recreate the group with the proper shapes.
    try:
        compatible = [group[key].shape == shape for key, shape in data_model.items()]
    except KeyError:
        compatible = (False,)
    compatible = all(compatible)
    if (not compatible) or reset:
        del f[group_name]
        init_h5_group(f, group_name, data_model)
    group = f[group_name]

    return group


def resize_h5_group(group, max_cycles):
    """Increase size of first dimension of datasets in the given group."""
    for key, dataset in group.items():
        # No need to resize scalar datasets
        if dataset.shape == ():
            continue
        new_shape = list(dataset.shape).copy()
        new_shape[0] = max_cycles
        dataset.resize(new_shape)



# [1] https://gaussian.com/uvvisplot/

import dataclasses
from pathlib import Path
from typing import Tuple

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray

from pysisyphus.constants import AU2J, C, M_E, NA, PLANCK, AU2EV


# Computation of prefactor from Gaussian whitepaper
Q_E_ESU = 4.803204e-10  # Crazy unit, cm**1.5 g**0.5 s⁻¹
C_CM = C * 100  # Speed of light in cm/s
M_E_G = M_E * 1000  # Electron mass in gram
NM2CM = 1e7  # Nanometer to centimeter
PREFACTOR = (  # Prefactor in eq. (5) of [1]
    np.sqrt(np.pi) * NA * Q_E_ESU**2 / (1e3 * np.log(10) * C_CM**2 * M_E_G)
) / NM2CM
_04EV = 0.4 / AU2EV  # in Hartree
# Factor used in converting energy in Hartree to wavelength
_AU2NM = PLANCK * C * 1e9 / AU2J


@dataclasses.dataclass
class Spectrum:
    exc_ens: NDArray[float]
    exc_ens_nm: NDArray[float]
    fosc: NDArray[float]
    nm: NDArray[float]
    epsilon: NDArray[float]

    def plot(self, **kwargs):
        plot_spectrum(self.nm, self.epsilon, self.exc_ens_nm, self.fosc, **kwargs)

    def save(self, fn):
        act_fn = Path(fn).with_suffix(".npz")
        data = dataclasses.asdict(self)
        np.savez(act_fn, **data)
        return act_fn

    @staticmethod
    def load(fn):
        handle = np.load(fn)
        kwargs = {key: handle[key] for key in handle.files}
        return Spectrum(**kwargs)

    @staticmethod
    def from_orca(wf_fn, cis_fn, log_fn, **kwargs):
        from pysisyphus.calculators.ORCA import get_exc_ens_fosc

        exc_ens, fosc = get_exc_ens_fosc(wf_fn, cis_fn, log_fn)
        return spectrum_from_ens_fosc(exc_ens, fosc, **kwargs)


def au2nm(au):
    return _AU2NM / au


def get_grid(resolution, exc_ens, padding, min_, max_):
    from_ = min(min_, int(exc_ens.min() - padding))
    to_ = max(max_, int(exc_ens.max() + padding))
    return np.arange(from_, to_ + resolution, step=resolution)


def homogeneous_broadening(
    exc_ens: NDArray[float],
    osc_strengths: NDArray[float],
    resolution: float = 0.5,
    stddev: float = _04EV,
    from_to=None,
) -> Tuple[NDArray[float], NDArray[float]]:
    """Homogeneous broadening of stick spectra as outlined in Gaussian
    whitepaper [1].

    σ = 0.0147 au corresponds to about 0.4 eV. The function yields molar
    extinction coefficients in l mol cm⁻¹."""
    exc_ens_nm = au2nm(exc_ens)
    if from_to is None:
        from_to = np.array((exc_ens_nm[0], exc_ens_nm[-1]))
    nm = get_grid(resolution, from_to, padding=100, min_=300, max_=900)
    stddev_nm = au2nm(stddev)

    quot = stddev_nm * (1 / nm[None, :] - (1 / exc_ens_nm[:, None]))
    exp_ = np.exp(-(quot**2))
    epsilon = PREFACTOR * osc_strengths[:, None] * stddev_nm * exp_
    epsilon = epsilon.sum(axis=0)
    return nm, epsilon


def spectrum_from_ens_fosc(exc_ens, fosc, **kwargs):
    exc_ens_nm = au2nm(exc_ens)
    nm, epsilon = homogeneous_broadening(exc_ens, fosc, **kwargs)
    spectrum = Spectrum(
        exc_ens=exc_ens,
        exc_ens_nm=exc_ens_nm,
        fosc=fosc,
        nm=nm,
        epsilon=epsilon,
    )
    return spectrum


def homogeneous_broadening_eV(
    exc_ens: NDArray[float],
    osc_strengths: NDArray[float],
    resolution: float = 0.05,
    stddev: float = 0.4,
    from_to=None,
) -> NDArray[float]:
    """
    Homogeneous broadening of stick spectra as outlined in eq. (5) and (6)
    of https://doi.org/10.1063/1.4948471.

    Parameters
    ----------
    exc_ens
        Excitation energies in Hartree.
    osc_strengths
        Unitless oscillator strengths.
    resolution
        Resolution of broadened spectra in eV.
    stddev
        Standard deviation, sigma in eq. (6).
    from_to
        Limits for abscissa of spectrum in eV.

    Returns
    -------
    eV
        Spectral grid in eV.
    epsilon
        Molar extinction coefficient.
    """
    exc_ens_eV = exc_ens * AU2EV
    if from_to is None:
        from_to = np.array((exc_ens_eV[0], exc_ens_eV[-1]))
    eV = get_grid(resolution, from_to, padding=2, min_=0, max_=5)
    epsilon = (osc_strengths / (3.7922e33 * 4 * 2.926e-39 * np.sqrt(np.pi) * stddev))[
        None, :
    ] * np.exp(-(((eV[None, :] - exc_ens_eV) / stddev) ** 2))
    epsilon = epsilon.sum(axis=0)
    return eV, epsilon


def plot_spectrum(nm, epsilon, exc_ens_nm=None, fosc=None, show=False):
    fig, ax = plt.subplots()
    ax.plot(nm, epsilon)
    ax.set_ylabel(r"$\epsilon$")
    ax.set_xlabel(r"$\lambda$ / nm")
    axs = [
        ax,
    ]

    if (exc_ens_nm is not None) and (fosc is not None):
        ax2 = ax.twinx()
        ax2.stem(exc_ens_nm, fosc, "r", markerfmt=" ", basefmt=" ")
        ax2.set_ylabel("fosc")
        axs.append(ax2)
    fig.tight_layout()
    if show:
        plt.show()

    return fig, axs



from collections import namedtuple
import logging

import numpy as np

from pysisyphus.config import T_DEFAULT
from pysisyphus.constants import FORCE2ACC
from pysisyphus.dynamics.helpers import (
    kinetic_energy_from_velocities,
    temperature_for_kinetic_energy,
    energy_forces_getter_closure,
    kinetic_energy_for_temperature,
    remove_com_velocity,
)
from pysisyphus.dynamics.thermostats import csvr_closure, csvr_closure_2, berendsen_closure
from pysisyphus.dynamics.rattle import rattle_closure
from pysisyphus.helpers import check_for_end_sign
from pysisyphus.helpers_pure import log
from pysisyphus.io.hdf5 import get_h5_group

logger = logging.getLogger("dynamics")

MDResult = namedtuple(
    "MDResult",
    "coords t_ps step terminated T E_tot",
)

THERMOSTATS = {
    "csvr": csvr_closure,
    "csvr_2": csvr_closure_2,
    "berendsen": berendsen_closure,
}


def get_data_model(atoms, dump_steps):
    coord_size = 3 * len(atoms)
    _1d = (dump_steps,)
    _2d = (dump_steps, coord_size)

    data_model = {
        "cart_coords": _2d,
        "step": _1d,
        "energy_tot": _1d,
        "energy_pot" : _1d,
        "energy_kin": _1d,
        "energy_conserved": _1d,
        "T": _1d,
        "T_avg": _1d,
        "velocity": _2d,
    }

    return data_model


def md(
    geom,
    v0,
    steps,
    dt,
    remove_com_v=True,
    thermostat=None,
    T=T_DEFAULT,
    timecon=100,
    term_funcs=None,
    constraints=None,
    constraint_kwargs=None,
    gaussians=None,
    verbose=True,
    print_stride=50,
    dump_stride=None,
    h5_group_name="run",
):
    """Velocity verlet integrator.

    Parameters
    ----------
    geom : Geometry
        The system for which the dynamics are to be run.
    v0 : np.array, floats
        Initial velocities in Bohr/fs.
    steps : float
        Number of simulation steps.
    dt : float
        Timestep in fs.
    remove_com_v : bool, default=True
        Remove center-of-mass velocity.
    thermostat : str, optional, default None
        Which and whether to use a thermostat.
    T : float, optional, default=None
        Desired temperature in thermostated runs.
    timecon : float
        Timeconsanst of the thermostat in fs.
    term_funcs : dict, optional
        Iterable of functions that are called with the atomic
        coordinates in every MD cycle and result in termination
    constraints : 2d iterable, optional
        2D iterable containing atom indices describing constrained
        bond lengths.
        of the MD integration when they evaluate to true.
    constraint_kwargs : dict, optional
        Keyword arguments for the constraint algorithm.
    gaussians: list, optional, default=None
        List of Gaussians to be used in a metadynamics run.
    verbose : bool, default=True
        Do additional printing when True.
    print_stride : int, default=50
        Report every n-th step.
    dump_stride : int, default=None
        If given, MD data will be dumped to a HDF5 file every n-th step.
    h5_group_name = str, optional
        Name of the HDF5 group, used for dumping.
    """

    assert geom.coord_type == "cart"

    if term_funcs is None:
        term_funcs = dict()

    # if dump_stride:
    #     dump_steps = steps // dump_stride
    #     data_model = get_data_model(geom.atoms, dump_steps)
    #     h5_group = get_h5_group("md.h5", h5_group_name, data_model, reset=True)
    #     h5_group.attrs["atoms"] = geom.atoms
    #     h5_group.attrs["dt"] = dt
    #     h5_group.attrs["steps"] = steps
    #     h5_group.attrs["masses"] = geom.masses
    #     h5_group.attrs["T_target"] = T

    if verbose:
        t_ps = steps * dt * 1e-3  # Total simulation time
        print(f"Doing {steps} steps of {dt:.4f} fs for a total of {t_ps:.2f} ps.\n")

    energy_forces_getter = energy_forces_getter_closure(geom)

    if gaussians is None:
        gaussians = tuple()
    gau_centers = list()
    gau_center_num = np.zeros(len(gaussians), dtype=int)
    for _, gaussian, gau_stride in gaussians:
        num_centers = steps // gau_stride
        gau_centers.append(np.zeros(num_centers))

    def update_gaussians(step, coords):
        for i, (gau_key, gaussian, gau_stride) in enumerate(gaussians):
            if (step + 1) % gau_stride == 0:
                new_center = gaussian.colvar.value(coords.reshape(-1, 3))
                gau_centers[i][gau_center_num[i]] = new_center
                gau_center_num[i] += 1
                log(
                    logger,
                    f"Added {gau_center_num[i]: >6d}. '{gau_key}' Gaussian in step {step}.",
                )
                gaussian.dump(step, gaussian.s, gaussian.w, new_center)

    def gaussian_wrapper(coords):
        E_pot, forces = energy_forces_getter(geom.coords)
        for i, (_, gaussian, _) in enumerate(gaussians):
            center_num = gau_center_num[i]
            gau_pot, gau_grad = gaussian.eval(coords, gau_centers[i][:center_num])
            # print("\tgau_pot", gau_pot)
            gau_forces = -gau_grad.flatten()
            E_pot += gau_pot
            forces += gau_forces
        return E_pot, forces

    if constraint_kwargs is None:
        constraint_kwargs = dict()

    # Fixed degrees of freedom
    fixed_dof = 0
    if remove_com_v:
        print("Removing center-of-mass velocity.")
        fixed_dof += 3
    constrained_md = constraints is not None
    # Get RATTLE function from closure for constrained MD
    if constrained_md:
        fixed_dof += len(constraints)
        rattle = rattle_closure(
            geom,
            constraints,
            dt,
            energy_forces_getter=energy_forces_getter,
            **constraint_kwargs,
        )
    print(f"Fixed degrees of freedom: {fixed_dof}")
    dof = len(geom.coords) - fixed_dof

    if thermostat is not None:
        sigma = kinetic_energy_for_temperature(len(geom.atoms), T, fixed_dof=fixed_dof)
        thermo_func = THERMOSTATS[thermostat](sigma, dof, dt=dt, tau=timecon)
    # In amu
    masses = geom.masses
    masses_rep = geom.masses_rep

    x = geom.cart_coords
    # v is given in Bohr/fs
    v = v0
    a_prev = np.zeros_like(x)
    xs = list()
    Ts = list()
    E_tots = list()

    E_pot, forces = energy_forces_getter(geom.coords)

    t_cur = 0
    terminate = False
    terminate_key = None
    T_avg = 0
    log(logger, f"Running MD with Δt={dt:.2f} fs for {steps} steps.")
    print()
    thermo_corr = 0.0
    for step in range(steps):
        xs.append(x.copy())

        E_kin = kinetic_energy_from_velocities(masses, v.reshape(-1, 3))
        T = temperature_for_kinetic_energy(len(masses), E_kin, fixed_dof=fixed_dof)
        T_avg += T
        Ts.append(T)
        E_tot = E_pot + E_kin
        E_tots.append(E_tot)
        E_conserved = E_pot + E_kin - thermo_corr

        status_msg = (
            f"Step {step:06d}  {t_cur*1e-3: >6.2f} ps  E={E_tot: >8.6f} E_h  "
            f"T={T: >8.2f} K <T>={T_avg/(step+1): >8.2f} K"
        )
        if step % print_stride == 0:
            log(logger, status_msg)
            if verbose:
                print(status_msg)

        # if dump_stride and (step % dump_stride == 0):
        #     ind = step // dump_stride
        #     h5_group["step"][ind] = step
        #     h5_group["cart_coords"][ind] = x
        #     h5_group["velocity"][ind] = v
        #     h5_group["energy_tot"][ind] = E_tot
        #     h5_group["energy_pot"][ind] = E_pot
        #     h5_group["energy_kin"][ind] = E_kin
        #     h5_group["energy_conserved"][ind] = E_conserved
        #     h5_group["T"][ind] = T
        #     h5_group["T_avg"][ind] = T_avg / (step + 1)

        if thermostat:
            alpha = thermo_func(E_kin)
            thermo_corr += (alpha**2 - 1) * E_kin
            v *= alpha

        update_gaussians(step, x)

        # RATTLE algorithm
        if constrained_md:
            x, v, E_pot, forces = rattle(x, v, forces)
        # Simple Velocity-Verlet integration
        else:
            # E_pot, forces = energy_forces_getter(geom.coords)
            E_pot, forces = gaussian_wrapper(geom.coords)
            # Acceleration, convert from Hartree / (Bohr * amu) to Bohr/fs²
            a = forces / masses_rep * FORCE2ACC
            v += 0.5 * (a + a_prev) * dt

            if remove_com_v:
                v = remove_com_velocity(v.reshape(-1, 3), masses).flatten()
            # v*dt = Bohr/fs * fs -> Bohr
            # a*dt**2 = Bohr/fs² * fs² -> Bohr
            x += v * dt + 0.5 * a * dt ** 2
            a_prev = a

        # Update coordinates
        geom.coords = x

        for name, func in term_funcs.items():
            if func(x.reshape(-1, 3)):
                terminate = True
                terminate_key = name
                break
        if terminate:
            log(logger, f"Termination function '{name}' evaluted to True. Breaking.")
            break

        if check_for_end_sign():
            break

        # Advance time
        t_cur += dt
    log(logger, "")

    md_result = MDResult(
        coords=np.array(xs),
        t_ps=t_cur * 1e-3,
        step=step,
        terminated=terminate_key,
        T=np.array(Ts),
        E_tot=np.array(E_tots),
    )

    return md_result



from pathlib import Path
from time import time

import numpy as np

from pysisyphus.config import T_DEFAULT, p_DEFAULT
from pysisyphus.constants import AU2KJPERMOL
from pysisyphus.helpers_pure import highlight_text, standard_state_corr
from pysisyphus.TablePrinter import TablePrinter
from pysisyphus.thermo import print_thermoanalysis


def do_endopt_ts_barriers(
    ts_geom,
    left_geoms,
    right_geoms=None,
    left_fns=None,
    right_fns=None,
    do_thermo=False,
    T=T_DEFAULT,
    p=p_DEFAULT,
    calc_getter=None,
    solv_calc_getter=None,
    do_standard_state_corr=False,
):
    print(highlight_text("Barrier heights after end optimization(s)"))
    print()

    # Only allow standard state correction when solvent calculator is specified.
    do_ssc = do_standard_state_corr and (solv_calc_getter is not None)
    ssc = standard_state_corr(T=T, p=p) if do_ssc else 0.0

    if right_geoms is None:
        right_geoms = []

    def set_fns(fns, geoms):
        if fns is None:
            fns = ["" for geom in geoms]
        else:
            fns = [Path(fn).name for fn in fns]
        return fns
    left_fns = set_fns(left_fns, left_geoms)
    right_fns = set_fns(right_fns, left_geoms)
    ts_fn = "TS"

    ts_atom_num = len(ts_geom.atoms)

    def drop_total_geom(geoms):
        atom_nums = [len(geom.atoms) for geom in geoms]
        tot_atom_num = sum(atom_nums)
        total_geom = None
        if tot_atom_num > ts_atom_num:
            total_ind = atom_nums.index(max(atom_nums))
            total_geom = geoms[total_ind]
            geoms = [geom for i, geom in enumerate(geoms) if i != total_ind]
        return geoms, total_geom

    left_geoms, left_total_geom = drop_total_geom(left_geoms)
    right_geoms, right_total_geom = drop_total_geom(right_geoms)

    def tot_atom_num(geoms):
        return sum([len(geom.atoms) for geom in geoms])

    left_atom_num = tot_atom_num(left_geoms)
    assert left_atom_num == ts_atom_num, (
        f"Atom number mismatch between TS ({ts_atom_num} atoms) and "
        f"left side ({left_atom_num} atoms)! Aborting barrier calculation"
    )
    right_tot_atom_num = tot_atom_num(right_geoms)
    assert (right_tot_atom_num == ts_atom_num) or (right_tot_atom_num == 0)

    def get_energy(geom, base_name):
        try:
            geom.energy
        except AttributeError:
            try:
                geom.set_calculator(calc_getter(base_name=base_name))
            except TypeError:
                raise Exception(
                    f"Energy isn't set at '{base_name}' geometry "
                    "and no 'calc_getter' was supplied!\nPlease either set "
                    "the desired quantities (energy/(Hessian)) or supply "
                    "a 'calc_getter' to calculate them."
                )
        return geom.energy

    def get_energies(geoms, base_name):
        return [
            get_energy(geom, f"{base_name}_{i:02d}") for i, geom in enumerate(geoms)
        ]

    en_key = "energy"
    ts_energy = get_energy(ts_geom, "ts")
    left_energies = get_energies(left_geoms, "left")
    right_energies = get_energies(right_geoms, "right")

    def zeros(geoms):
        return np.zeros(len(geoms))

    # Gibbs free energies
    if do_thermo:
        en_key = "free energy"

        def get_thermo(geom, title, is_ts=False):
            thermo = geom.get_thermoanalysis(geom, T=T, p=p)
            print_thermoanalysis(thermo, geom=geom, is_ts=is_ts, level=1, title=title)
            print()
            return thermo

        ts_thermo = get_thermo(ts_geom, "TS", is_ts=True)
        ts_dG = ts_thermo.dG
        left_thermos = [get_thermo(geom, fn) for geom, fn in zip(left_geoms, left_fns)]
        left_dGs = [thermo.dG for thermo in left_thermos]
        right_thermos = [
            get_thermo(geom, fn) for geom, fn in zip(right_geoms, right_fns)
        ]
        right_dGs = [thermo.dG for thermo in right_thermos]
    else:
        ts_dG = 0.0
        left_dGs = zeros(left_geoms)
        right_dGs = zeros(right_geoms)

    def get_solv_correction(geom, fn=None, name=None):
        fn = str(fn)
        if fn is not None:
            infix = f"'{Path(fn).name: >40s}'"
        else:
            atom_num = len(geom.atoms)
            infix = f"{atom_num} atoms"

        print(f"Solvated calculation for {infix} ...", end="")
        start = time()
        solv_calc_kwargs = {}
        if name is not None:
            solv_calc_kwargs["base_name"] = f"solv_{name}"
        solv_energy = solv_calc_getter(**solv_calc_kwargs).get_energy(
            geom.atoms, geom.cart_coords
        )["energy"]
        dur = time() - start
        print(f" finished in {dur:.1f} s.")
        # Add standard state correction. ssc will be 0.0 if, disabled.
        return solv_energy, (solv_energy - geom.energy) + ssc

    def get_solv_corrections(geoms, fns, name):
        return zip(
            *[get_solv_correction(geom, fn, name) for geom, fn in zip(geoms, fns)]
        )

    if solv_calc_getter is not None:
        print(highlight_text("Calculations with solvent model", level=1) + "\n")
        ts_solv_energy, ts_solv_corr = get_solv_correction(ts_geom, ts_fn, "ts")
        left_solv_energies, left_solv_corrs = get_solv_corrections(
            left_geoms, left_fns, "left"
        )
        right_solv_energies, right_solv_corrs = get_solv_corrections(
            right_geoms, right_fns, "right"
        )
        print()
    # Use zeros, so we can add the term later.
    else:
        ts_solv_corr = 0.0
        left_solv_corrs = zeros(left_geoms)
        right_solv_corrs = zeros(right_geoms)

    # Calculate total contributions for both sides
    left_dG = sum(left_dGs)
    right_dG = sum(right_dGs)
    left_solv_corr = sum(left_solv_corrs)
    right_solv_corr = sum(right_solv_corrs)

    ts_energy_corr = ts_energy + ts_solv_corr + ts_dG
    left_energy_corr = sum(left_energies) + left_dG + left_solv_corr
    right_energy_corr = sum(right_energies) + right_dG + right_solv_corr
    energies_corr = [left_energy_corr, ts_energy_corr]
    # TS is always only 1 geometry
    geom_nums = [len(left_geoms), 1]

    fns = ["Left", "TS"]
    if right_geoms:
        energies_corr.append(right_energy_corr)
        geom_nums.append(len(right_geoms))
        fns.append("Right")
    max_len = max(len(s) for s in fns)

    def print_table(table, header, width=17):
        col_fmts = ["str"] + (len(header) - 1) * ["float"]
        tp = TablePrinter(header, col_fmts, width=width, sub_underline=False)
        tp.print_header()
        for fn, row in zip(all_fns, table.T):
            fn = str(fn)  # fns may be PosixPath etc.
            fn_cut = fn[:4] + ".." + fn[-11:] if (len(fn) > width) else fn
            tp.print_row((fn_cut, *row))
        print()

    all_fns = left_fns + [ts_fn] + right_fns
    # Electronic energies
    all_energies = np.concatenate((left_energies, [ts_energy], right_energies))
    to_stack = [all_energies]
    gas_titles = ["File", "E_el"]
    if do_thermo:
        # Thermochemical corrections
        all_dGs = np.concatenate((left_dGs, [ts_dG], right_dGs))
        # Free Gibbs energies in the gas phase
        all_Gs_gas = all_energies + all_dGs
        to_stack += [all_dGs, all_Gs_gas]
        gas_titles += ["dG_gas", "G_gas"]

    gas_table = np.stack(to_stack)
    print("All tabulated quantities are given in au.\n")
    print_table(gas_table, gas_titles)
    full_titles = gas_titles

    if solv_calc_getter is not None:
        all_solv_energies = np.concatenate(
            (left_solv_energies, [ts_solv_energy], right_solv_energies)
        )
        all_solv_corrs = np.concatenate(
            (left_solv_corrs, [ts_solv_corr], right_solv_corrs)
        )
        to_stack = [all_energies, all_solv_energies, all_solv_corrs]
        solv_titles = ["File", "E_el", "E_solv", "dG_solv"]
        if do_thermo:
            all_Gs_sol = all_Gs_gas + all_solv_corrs
            to_stack += [all_Gs_sol]
            solv_titles += ["G_sol"]
        solv_table = np.stack(to_stack)
        if do_ssc:
            print("dG_solv includes a correction for change of standard state.\n")
        print_table(solv_table, solv_titles)
        full_titles += solv_titles[2:]

    energies_corr = np.array(energies_corr)
    energies_corr -= energies_corr.min()
    min_ind = energies_corr.argmin()
    energies_corr *= AU2KJPERMOL

    print(highlight_text("Barriers", level=1))
    print()
    print(f"Temperature: {T:.2f} K")
    print(f"Pressure: {p:.1f} Pa\n")

    print("Corrections:")
    print(f"Change of standard-state: {do_ssc}, {ssc*AU2KJPERMOL:.2} kJ mol⁻¹")
    print(f"                 Solvent: {solv_calc_getter is not None}")
    print(f"         Thermochemistry: {do_thermo}")
    print()

    def print_geoms_fns(geoms, fns):
        for i, (geom, fn) in enumerate(zip(geoms, fns)):
            fn_name = Path(fn).name
            print(f"\t{i}: {fn_name} ({geom}, {len(geom.atoms)} atoms)")

    def get_geom_key(geoms):
        return "geometry" if len(geoms) == 1 else "geometries"

    print(f"Left {get_geom_key(left_geoms)}:")
    print_geoms_fns(left_geoms, left_fns)
    print("TS geometry:")
    print_geoms_fns((ts_geom,), ("",))
    if right_geoms:
        print(f"Right {get_geom_key(right_geoms)}:")
        print_geoms_fns(right_geoms, right_fns)
    print()

    print(f"Minimum {en_key} of {energies_corr[min_ind]} kJ mol⁻¹ at '{fns[min_ind]}'.")
    print()
    for fn, en, gnum in zip(fns, energies_corr, geom_nums):
        geom_str = "geometry" if gnum == 1 else "geometries"
        is_sum = " " if gnum == 1 else "Σ"
        print(f"\t{is_sum}{fn:>{max_len}s}: {en:>8.2f} kJ mol⁻¹ ({gnum} {geom_str})")
    print()



# bond_changes.py

"""
bond_changes — Bond-change detection and reporting utilities
====================================================================

Usage (API)
-----
    from <package>.bond_changes import compare_structures, summarize_changes

Examples::
    >>> result = compare_structures(geom_reactant, geom_product, device="cpu")
    >>> print(summarize_changes(geom_product, result))
    Bond formed (1)
      C1-O2 1.50 Å --> 1.36 Å

Description
-----
This module compares two molecular geometries with identical atom types and ordering and reports covalent bonds that are formed
or broken between the structures. Bond perception uses element-specific covalent radii and configurable tolerances; distances are computed with PyTorch on CPU or CUDA.

Algorithm (core logic):
- Inputs: `geom1`, `geom2` with attributes `atoms: Iterable[str]` and `coords3d: (N, 3) float` (in Bohr in pysisyphus). Atoms must match exactly (`assert geom1.atoms == geom2.atoms`).
- Per-element radii: from `pysisyphus.elem_data.COVALENT_RADII`.
- Threshold per pair (i, j): `T_cov = bond_factor * (r_i + r_j)`; conservative margin `eps = margin_fraction * T_cov`.
- Bond adjacency in a geometry: `A = (D <= T_cov - eps)` evaluated only for `i < j` (upper triangle).
- Change gating: only pairs with `|D2 - D1| >= delta_fraction * T_cov` are considered.
- Classification:
    - Formed: `(~A1) & A2 & need_change`
    - Broken:  `A1 & (~A2) & need_change`
- Distances: pairwise matrices `D1`, `D2` via `torch.cdist`.

Public API:
- `compare_structures(geom1, geom2, device='cuda', bond_factor=1.20, margin_fraction=0.05, delta_fraction=0.05) -> BondChangeResult`
  Detects formed and broken covalent bonds. Returns sets of zero-based index pairs and (by default) the full distance matrices (`numpy.ndarray`) in the same units as the inputs (Bohr in pysisyphus).
- `summarize_changes(geom, result, one_based: bool = True) -> str`
  Builds a human-readable report:
  - Sections: “Bond formed” and “Bond broken” (with counts).
  - Lines formatted as `ElemI-ElemJ` with atom indices (1-based by default, e.g., `C1-O2`).
  - If `result.distances_1/2` are present, prints bond lengths as `D1 Å --> D2 Å`, converting from Bohr using `pysisyphus.constants.BOHR2ANG`.
- Helper utilities (internal):
  - `_resolve_device(device: str) -> torch.device`: chooses the requested device; falls back to CPU with a warning if unavailable.
  - `_element_arrays(atoms) -> (elems, cov_radii)`: normalizes element symbols and looks up covalent radii.
  - `_upper_pairs_from_mask(mask) -> Set[Tuple[int, int]]`: returns index pairs where `mask` is True (assumes an upper-triangular mask).
  - `_bond_str(i, j, elems, one_based=True) -> str`: formats `ElemI-ElemJ` labels.
- Data container:
  - `BondChangeResult`:
    - `formed_covalent: Set[Tuple[int, int]]` — zero-based pairs for bonds formed.
    - `broken_covalent: Set[Tuple[int, int]]` — zero-based pairs for bonds broken.
    - `distances_1: Optional[np.ndarray]`, `distances_2: Optional[np.ndarray]` — square distance matrices (shape N×N).

Outputs (& Directory Layout)
-----
- No files or directories are created.
- `compare_structures` returns a `BondChangeResult` as described above.
- `summarize_changes` returns a multi-line string; typical headings:
  - `Bond formed (k):` followed by `ElemI-ElemJ : D1 Å --> D2 Å` (if distances available).
  - `Bond broken (m):` followed by lines in the same format.
  - If a set is empty, the section reads `None`.

Notes:
-----
- Units: In pysisyphus, `coords3d` are Bohr; the summary converts to Å with `BOHR2ANG`. If your inputs use different units, adjust accordingly.
- Atom identity & ordering must be identical between structures; otherwise the comparison is invalid.
- Only unique pairs with `i < j` are considered (upper triangle); indices in results are zero-based.
- The three tolerances control sensitivity:
  - `bond_factor` (default 1.20): global scaling of the covalent radii sum.
  - `margin_fraction` (default 0.05): conservative shrinkage of the bond cutoff to avoid borderline matches.
  - `delta_fraction` (default 0.05): minimum relative distance change required to count a bond event.
- Device selection: pass `'cpu'`, `'cuda'`, or `'cuda:0'` etc. If the requested device is not available or invalid, the code falls back to CPU and issues a `RuntimeWarning`. Computations use `float64` for stability and run under `torch.no_grad()`.
- The method detects binary bond formation/breakage; it does not estimate bond orders, angles, or multi-center bonding, and it ignores periodic boundary conditions.
- Numerical caveats: near-threshold pairs may toggle with small geometry noise; tune tolerances to your system size and sampling noise.
"""

from __future__ import annotations
from dataclasses import dataclass
from typing import Iterable, Tuple, Set, List, Optional, Dict, Any

import warnings
import torch
import numpy as np

from pysisyphus.elem_data import (
    COVALENT_RADII as CR,
)
from pysisyphus.constants import BOHR2ANG  # Convert Bohr distances to Ångström for reporting

Pair = Tuple[int, int]


@dataclass
class BondChangeResult:
    formed_covalent: Set[Pair]
    broken_covalent: Set[Pair]
    distances_1: Optional[np.ndarray] = None
    distances_2: Optional[np.ndarray] = None


def _upper_pairs_from_mask(mask: torch.Tensor) -> Set[Pair]:
    idx = torch.nonzero(mask, as_tuple=False).detach().cpu().numpy()
    return set(map(tuple, idx))


def _element_arrays(atoms: Iterable[str]) -> Tuple[List[str], np.ndarray]:
    elems = [a.capitalize() for a in atoms]
    cov = np.array([CR[a.lower()] for a in elems], dtype=float)
    return elems, cov


def _resolve_device(device: str) -> torch.device:
    dev_str = (device or "cpu").lower()
    if dev_str.startswith("cuda"):
        if torch.cuda.is_available():
            try:
                _ = torch.device(dev_str)
                return torch.device(dev_str)
            except Exception:
                warnings.warn(
                    f"Requested device '{device}' is not available. Falling back to CPU.",
                    RuntimeWarning,
                )
                return torch.device("cpu")
        else:
            warnings.warn(
                "CUDA is not available. Falling back to CPU.",
                RuntimeWarning,
            )
            return torch.device("cpu")
    try:
        return torch.device(dev_str)
    except Exception:
        warnings.warn(
            f"Requested device '{device}' is not recognized. Falling back to CPU.",
            RuntimeWarning,
        )
        return torch.device("cpu")


@torch.no_grad()
def compare_structures(
    geom1,
    geom2,
    device: str = "cuda",
    bond_factor: float = 1.20,
    margin_fraction: float = 0.05,
    delta_fraction: float = 0.05,
) -> BondChangeResult:

    assert geom1.atoms == geom2.atoms, "Atom types and ordering must be identical."
    N = len(geom1.atoms)

    elems, cov_np = _element_arrays(geom1.atoms)
    dev = _resolve_device(device)

    dtype = torch.float64
    R1 = torch.as_tensor(geom1.coords3d, dtype=dtype, device=dev)
    R2 = torch.as_tensor(geom2.coords3d, dtype=dtype, device=dev)
    cov = torch.as_tensor(cov_np, dtype=dtype, device=dev)

    T_cov = bond_factor * (cov[:, None] + cov[None, :])
    eps_cov = margin_fraction * T_cov

    D1 = torch.cdist(R1, R1)
    D2 = torch.cdist(R2, R2)

    up = torch.triu(torch.ones((N, N), dtype=torch.bool, device=dev), diagonal=1)

    A1 = (D1 <= (T_cov - eps_cov)) & up
    A2 = (D2 <= (T_cov - eps_cov)) & up

    dD = D2 - D1
    need_change = (dD.abs() >= (delta_fraction * T_cov)) & up

    formed_cov_mask = (~A1) & A2 & need_change
    broken_cov_mask = A1 & (~A2) & need_change

    formed_covalent = _upper_pairs_from_mask(formed_cov_mask)
    broken_covalent = _upper_pairs_from_mask(broken_cov_mask)

    return BondChangeResult(
        formed_covalent=formed_covalent,
        broken_covalent=broken_covalent,
        distances_1=D1.detach().cpu().numpy(),
        distances_2=D2.detach().cpu().numpy(),
    )


def _bond_str(i: int, j: int, elems: List[str], one_based: bool = True) -> str:
    ii = i + 1 if one_based else i
    jj = j + 1 if one_based else j
    return f"{elems[i]}{ii}-{elems[j]}{jj}"


def summarize_changes(geom, result: BondChangeResult, one_based: bool = True) -> str:
    """
    List bond formations and dissociations and report bond-length changes in Å.
    """
    elems = [a.capitalize() for a in geom.atoms]
    lines: List[str] = []

    # Use distance matrices (Bohr) converted to Å when available
    D1 = result.distances_1
    D2 = result.distances_2
    have_lengths = (
        isinstance(D1, np.ndarray)
        and isinstance(D2, np.ndarray)
        and D1.shape == D2.shape
    )

    def _len_str(i: int, j: int) -> str:
        if not have_lengths:
            return ""
        # ``coords3d`` is given in Bohr; convert to Å
        d1 = float(D1[i, j]) * BOHR2ANG
        d2 = float(D2[i, j]) * BOHR2ANG
        return f" : {d1:.3f} Å --> {d2:.3f} Å"

    def pairs_to_lines(title: str, pairs: Set[Pair]):
        if not pairs:
            lines.append(f"{title}: None")
            return
        lines.append(f"{title} ({len(pairs)}):")
        for i, j in sorted(pairs):
            lines.append(f"  - {_bond_str(i, j, elems, one_based)}{_len_str(i, j)}")

    pairs_to_lines("Bond formed", result.formed_covalent)
    pairs_to_lines("Bond broken", result.broken_covalent)

    return "\n".join(lines)


def has_bond_change(
    geom_start,
    geom_end,
    bond_cfg: Dict[str, Any],
    *,
    one_based: bool = True,
) -> Tuple[bool, str]:
    """
    Convenience wrapper to check if bond changes occur between two geometries.

    Parameters
    ----------
    geom_start : Geometry
        Starting geometry.
    geom_end : Geometry
        Ending geometry.
    bond_cfg : Dict[str, Any]
        Configuration dict with keys: 'device', 'bond_factor', 'margin_fraction', 'delta_fraction'.
    one_based : bool
        Use 1-based atom indices in summary (default True).

    Returns
    -------
    Tuple[bool, str]
        (has_changes, summary) where has_changes is True if any bonds formed or broken.
    """
    device = bond_cfg.get("device", "cuda")
    bond_factor = bond_cfg.get("bond_factor", 1.20)
    margin_fraction = bond_cfg.get("margin_fraction", 0.05)
    delta_fraction = bond_cfg.get("delta_fraction", 0.05)

    result = compare_structures(
        geom_start,
        geom_end,
        device=device,
        bond_factor=bond_factor,
        margin_fraction=margin_fraction,
        delta_fraction=delta_fraction,
    )

    has_changes = bool(result.formed_covalent) or bool(result.broken_covalent)
    summary = summarize_changes(geom_start, result, one_based=one_based)

    return has_changes, summary



import argparse
import itertools as it
import sys

import matplotlib.pyplot as plt
import numpy as np
import yaml

from pysisyphus.constants import AU2EV
from pysisyphus.wavefunction.diabatization import dq_diabatization


def make_array(nstates, components, lines):
    if len(lines) == 0:
        return None

    arr = np.zeros((components, nstates, nstates))
    expect_diag = nstates
    expect_off_diag = sum(range(nstates))

    for line in lines:
        from_, to_, *props = line
        assert (from_ >= 0) and (to_ >= 0)
        assert (
            len(props) == components
        ), f"Expected line of length {components} but got '{line}'!"

        if from_ == to_:
            expect_diag -= 1
        elif from_ != to_:
            expect_off_diag -= 1

        arr[:, from_, to_] = arr[:, to_, from_] = props
    assert expect_diag == 0, "Some diagonal elements are missing!"
    assert expect_off_diag == 0, "Some off-diagonal elements are missing!"
    return arr


def dq_diabatization_from_run_dict(run_dict):
    adia_ens = np.array(run_dict["adiabatic_energies"], dtype=float)
    nstates = adia_ens.size

    # Dipole moments must be present
    dip_moms = make_array(nstates, 3, run_dict["dipoles"])
    # Quadrupole moments and electrostatic potential are optional.
    quad_moms = make_array(nstates, 1, run_dict.get("quadrupoles", list()))
    epots = make_array(nstates, 1, run_dict.get("epots", list()))

    kwargs = {}
    if "alpha" in run_dict:
        kwargs["alpha"] = run_dict["alpha"]
    if "beta" in run_dict:
        kwargs["beta"] = run_dict["beta"]

    return dq_diabatization(
        adia_ens, dip_moms, quad_moms=quad_moms, epots=epots, **kwargs
    )


def diabatize_path(adia_ens, dip_moms, tr_quad_moms=None, epots=None, **kwargs):
    nones = [None for _ in adia_ens]
    if tr_quad_moms is None:
        tr_quad_moms = nones
    if epots is None:
        epots = nones
    assert len(adia_ens) == len(dip_moms) == len(tr_quad_moms) == len(epots)

    for aens, dpm, qpm, epot in zip(adia_ens, dip_moms, tr_quad_moms, epots):
        yield dq_diabatization(aens, dpm, qpm, epot, **kwargs)


def plot_dia_res(dia_res, show=False):
    nstates = dia_res[0].nstates
    adia_ens = np.zeros((len(dia_res), nstates))
    dia_ens = np.zeros((len(dia_res), nstates))
    keys = list(it.combinations(range(nstates), 2))
    couplings = np.zeros((len(dia_res), len(keys)))
    for i, dr in enumerate(dia_res):
        adia_ens[i] = dr.adia_ens
        dia_ens[i] = dr.dia_ens
        for j, key in enumerate(keys):
            couplings[i, j] = dr.couplings[key]

    adia_min = adia_ens.min()
    adia_ens -= adia_min
    adia_ens *= AU2EV
    dia_ens -= adia_min
    dia_ens *= AU2EV
    couplings *= AU2EV

    fig, (ax0, ax1) = plt.subplots(nrows=2, sharex=True)
    for i, state in enumerate(adia_ens.T):
        ax0.plot(state, "o--", label=f"$V_{i}$")
    for i, state in enumerate(dia_ens.T):
        ax0.plot(state, "x-", label=f"$U_{i}$")
    ax0.legend()
    ax0.set_xlabel("Step")
    ax0.set_ylabel(r"$\Delta E$ / eV")

    # Couplings
    for i, cpls in enumerate(couplings.T):
        from_to = "".join([str(_) for _ in keys[i]])
        ax1.plot(cpls, "o-", label=f"$|D_{{{from_to}}}|$")
    ax1.axhline(0.0, ls="--", c="k")
    ax1.legend()

    fig.tight_layout()
    if show:
        plt.show()
    return fig, (ax0, ax1)


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("yaml")
    return parser.parse_args(args)


def run():
    args = parse_args(sys.argv[1:])

    yaml_fn = args.yaml
    with open(yaml_fn) as handle:
        run_dict = yaml.load(handle, Loader=yaml.SafeLoader)
    adia_labels = run_dict.pop("adiabatic_labels", None)
    unit = run_dict.pop("unit", "eV")
    dia_res = dq_diabatization_from_run_dict(run_dict)
    report = dia_res.render_report(adia_labels=adia_labels, unit=unit)
    print(report)


if __name__ == "__main__":
    run()



import abc
import types

import numpy as np
import autograd
import autograd.numpy as anp

from pysisyphus.intcoords.derivatives import dq_b, dq_a


class Colvar(metaclass=abc.ABCMeta):
    def __init__(self, force_agrad=False):
        try:
            getattr(self, "_gradient")
        except AttributeError:
            force_agrad = True

        # Set autograd gradient method, if no _gradient is not implemented
        # a autograd is forced.
        if force_agrad:
            grad_func = autograd.grad(self.value)

            def wrapped(self, c3d):
                return grad_func(c3d)

            self._gradient = types.MethodType(wrapped, self)
        # Store a flag to indicate use of autograd
        self.agrad = force_agrad

    @abc.abstractmethod
    def value(self, c3d):
        pass

    def gradient(self, c3d):
        return self._gradient(c3d)

    def eval(self, coords):
        c3d = coords.reshape(-1, 3)
        return self.value(c3d), self.gradient(c3d)

    def _wilson_gradient(self, func, c3d):
        """Gradient of primitive internal w.r.t. Cartesians."""
        grad = np.zeros_like(c3d)
        grad[self.indices] = func(*c3d[self.indices].flatten()).reshape(-1, 3)
        return grad

    def __str__(self):
        if hasattr(self, "indices"):
            str_ = f"{self.__class__.__name__}({self.indices})"
        else:
            str_ = super().__str__()
        return str_


class CVDistance(Colvar):
    def __init__(self, indices, **kwargs):
        self.indices = list(indices)
        self.i, self.j = self.indices
        super().__init__(**kwargs)

    def value(self, c3d):
        return anp.linalg.norm(c3d[self.i] - c3d[self.j])

    def _gradient(self, c3d):  # lgtm [py/attribute-shadows-method]
        return self._wilson_gradient(dq_b, c3d)


class CVBend(Colvar):
    def __init__(self, indices, **kwargs):
        self.indices = list(indices)
        # Bonded like
        # i---j <- central atom
        #     |
        #     k
        self.i, self.j, self.k = self.indices
        super().__init__(**kwargs)

    def value(self, c3d):
        u_dash = c3d[self.i] - c3d[self.j]
        v_dash = c3d[self.k] - c3d[self.j]
        u_norm = anp.linalg.norm(u_dash)
        v_norm = anp.linalg.norm(v_dash)
        u = u_dash / u_norm
        v = v_dash / v_norm
        rad = anp.arccos(anp.dot(u, v))
        return rad

    def _gradient(self, c3d):  # lgtm [py/attribute-shadows-method]
        return self._wilson_gradient(dq_a, c3d)


class CVTorsion(Colvar):
    def __init__(self, indices, **kwargs):
        self.indices = list(indices)
        # Bonded like
        # i--u--j
        #       |
        #       w
        #       |
        #       k--v--l
        self.i, self.j, self.k, self.l = self.indices
        super().__init__(**kwargs)

    def value(self, c3d):
        u_dash = c3d[self.i] - c3d[self.j]
        v_dash = c3d[self.l] - c3d[self.k]
        w_dash = c3d[self.k] - c3d[self.j]
        u_norm = anp.linalg.norm(u_dash)
        v_norm = anp.linalg.norm(v_dash)
        w_norm = anp.linalg.norm(w_dash)
        u = u_dash / u_norm
        v = v_dash / v_norm
        w = w_dash / w_norm
        phi_u = anp.arccos(anp.dot(u, w))
        phi_v = anp.arccos(-anp.dot(w, v))
        uxw = anp.cross(u, w)
        vxw = anp.cross(v, w)
        cos_dihed = anp.dot(uxw, vxw)/(anp.sin(phi_u)*anp.sin(phi_v))
        # Restrict cos_dihed to the allowed interval for arccos [-1, 1]
        cos_dihed = min(1, max(cos_dihed, -1))

        rad = anp.arccos(cos_dihed)
        if (rad != np.pi) and (anp.dot(vxw, u) < 0):
            rad *= -1
        return rad


COLVARS = {
    "distance": CVDistance,
    "bend": CVBend,
    "torsion": CVTorsion,
}


def get_colvar(key, kwargs):
    return COLVARS[key](**kwargs)



#!/usr/bin/env python3

# Johannes Steinmetzer 2020
# As describend in:
#   V. Bakken, T. Helgaker, J. Chem. Phys., 117, 20, 2002
# [1] https://aip.scitation.org/doi/abs/10.1063/1.1515483
# [2] https://doi.org/10.1002/(SICI)1096-987X(19960115)17:1<49::AID-JCC5>3.0.CO;2-0
# [3] TRANSITION-STATE OPTIMIZATION METHODS USING INTERNAL COORDINATES
#     Sandra Rabi, PhD Thesis

from collections import namedtuple
import itertools as it
import random
import string
import textwrap

from jinja2 import Template
import sympy as sym
from sympy import cse
from sympy.codegen.ast import Assignment
from sympy.printing.pycode import pycode, MpmathPrinter
from sympy.vector import CoordSys3D
from sympy.tensor.array.dense_ndim_array import ImmutableDenseNDimArray


FuncResult = namedtuple("FuncResult", "d0 d1 d2 f0 f1 f2")


def make_py_func(exprs, args=None, name=None, comment="", use_mpmath=False):
    if args is None:
        args = list()
    if name is None:
        name = "func_" + "".join(
            [random.choice(string.ascii_letters) for i in range(8)]
        )
    arg_strs = [arg.strip() for arg in args.split(",")]

    is_scalar = not isinstance(exprs, ImmutableDenseNDimArray)
    if is_scalar:
        repls, reduced = cse(exprs)
        reduced = reduced[0]
    else:
        if len(exprs.shape) == 2:
            exprs = it.chain(*exprs)
        repls, reduced = cse(list(exprs))

    print_func = pycode
    if use_mpmath:
        print_func = MpmathPrinter().doprint

    assignments = [Assignment(lhs, rhs) for lhs, rhs in repls]
    py_lines = [print_func(as_) for as_ in assignments]
    return_val = print_func(reduced)

    tpl = Template(
        """
    def {{ name }}({{ args }}):
        {% if comment %}
        \"\"\"{{ comment }}\"\"\"
        {% endif %}

        {% if use_mpmath %}
        {% for arg in arg_strs %}
        {{ arg }} = mpmath.mpf({{ arg }})
        {% endfor %}
        {% endif %}

        {% for line in py_lines %}
        {{ line }}
        {% endfor %}

        {% if is_scalar %}
        return {{ return_val }}
        {% else %}
        return np.array({{ return_val }}, dtype=np.float64)
        {% endif %}
    """,
        trim_blocks=True,
        lstrip_blocks=True,
    )

    rendered = textwrap.dedent(
        tpl.render(
            name=name,
            args=args,
            py_lines=py_lines,
            return_val=return_val,
            comment=comment,
            is_scalar=is_scalar,
            use_mpmath=use_mpmath,
            arg_strs=arg_strs,
        )
    ).strip()
    return rendered


# def make_deriv_funcs(base_expr, dx, args, names, comments):
def make_deriv_funcs(base_expr, dx, args, names, comment, use_mpmath=True):
    q_name, d1_name, d2_name = names
    q_comment, d1_comment, d2_comment = [
        comment + add
        for add in [
            "",
            ", first derivative wrt. Cartesians",
            ", 2nd derivative wrt. Cartesians",
        ]
    ]

    # Actual function
    print(f"\tmpmath: {use_mpmath}")
    print("\tFunction")
    q_func = make_py_func(
        base_expr, args=args, name=q_name, comment=q_comment, use_mpmath=use_mpmath
    )

    # First derivative
    print("\t1st derivative")
    deriv1 = sym.derive_by_array(base_expr, dx)
    deriv1_func = make_py_func(
        deriv1, args=args, name=d1_name, comment=d1_comment, use_mpmath=use_mpmath
    )

    # Second derivative
    print("\t2nd derivative")
    deriv2 = sym.derive_by_array(deriv1, dx)
    deriv2_func = make_py_func(
        deriv2, args=args, name=d2_name, comment=d2_comment, use_mpmath=use_mpmath
    )

    return FuncResult(
        # Expressions
        d0=base_expr,
        d1=deriv1,
        d2=deriv2,
        # Functions
        f0=q_func,
        f1=deriv1_func,
        f2=deriv2_func,
    )


def generate_wilson(generate=None, out_fn="derivatives.py", use_mpmath=False):
    m0, m1, m2, n0, n1, n2, o0, o1, o2, p0, p1, p2 = sym.symbols("m:3 n:3 o:3 p:3")

    # Coordinate system
    Sys = CoordSys3D("Sys")
    M = Sys.origin.locate_new("M", m0 * Sys.i + m1 * Sys.j + m2 * Sys.k)
    N = Sys.origin.locate_new("N", n0 * Sys.i + n1 * Sys.j + n2 * Sys.k)
    O = Sys.origin.locate_new("O", o0 * Sys.i + o1 * Sys.j + o2 * Sys.k)
    P = Sys.origin.locate_new("P", p0 * Sys.i + p1 * Sys.j + p2 * Sys.k)

    def bond():
        # Bond/Stretch
        U = M.position_wrt(N)
        q_b = U.magnitude()
        dx_b = (m0, m1, m2, n0, n1, n2)
        args_b = "m0, m1, m2, n0, n1, n2"
        func_result_b = make_deriv_funcs(
            q_b,
            dx_b,
            args_b,
            ("q_b", "dq_b", "d2q_b"),
            "Stretch",
            use_mpmath=use_mpmath,
        )
        return func_result_b

    def bend():
        # Bend/Angle
        U = M.position_wrt(O)
        V = N.position_wrt(O)
        q_a = sym.acos(U.dot(V) / (U.magnitude() * V.magnitude()))
        dx_a = (m0, m1, m2, o0, o1, o2, n0, n1, n2)
        args_a = "m0, m1, m2, o0, o1, o2, n0, n1, n2"
        func_result_a = make_deriv_funcs(
            q_a,
            dx_a,
            args_a,
            ("q_a", "dq_a", "d2q_a"),
            "Bend",
            use_mpmath=use_mpmath,
        )
        return func_result_a

    def bend2():
        # atan2 based Bend/Angle
        U = M.position_wrt(O)
        V = N.position_wrt(O)
        q_a2 = sym.atan2(U.cross(V).magnitude(), U.dot(V))
        dx_a2 = (m0, m1, m2, o0, o1, o2, n0, n1, n2)
        args_a2 = "m0, m1, m2, o0, o1, o2, n0, n1, n2"
        func_result_a = make_deriv_funcs(
            q_a2,
            dx_a2,
            args_a2,
            ("q_a2", "dq_a2", "d2q_a2"),
            "Bend2",
            use_mpmath=use_mpmath,
        )
        return func_result_a

    def dihedral():
        # Dihedral/Torsion
        U = M.position_wrt(O)
        W = P.position_wrt(O)
        V = N.position_wrt(P)
        U_ = U.normalize()
        W_ = W.normalize()
        V_ = V.normalize()
        phi_u = sym.acos(U_.dot(W_))
        phi_v = sym.acos(-W_.dot(V_))
        q_d = sym.acos(
            U_.cross(W_).dot(V_.cross(W_)) / (sym.sin(phi_u) * sym.sin(phi_v))
        )
        dx_d = (m0, m1, m2, o0, o1, o2, p0, p1, p2, n0, n1, n2)
        args_d = "m0, m1, m2, o0, o1, o2, p0, p1, p2, n0, n1, n2"
        func_result_d = make_deriv_funcs(
            q_d,
            dx_d,
            args_d,
            ("q_d", "dq_d", "d2q_d"),
            "Torsion",
            use_mpmath=use_mpmath,
        )
        return func_result_d

    def dihedral2():
        # atan2 based Dihedral/Torsion
        U1 = O.position_wrt(M)
        U2 = P.position_wrt(O)
        U3 = N.position_wrt(P)
        cross_U2U3 = U2.cross(U3)
        q_d2 = sym.atan2(
            (U2.magnitude() * U1).dot(cross_U2U3), cross_U2U3.dot(U1.cross(U2))
        )
        dx_d2 = (m0, m1, m2, o0, o1, o2, p0, p1, p2, n0, n1, n2)
        args_d2 = "m0, m1, m2, o0, o1, o2, p0, p1, p2, n0, n1, n2"
        func_result_d = make_deriv_funcs(
            q_d2,
            dx_d2,
            args_d2,
            ("q_d2", "dq_d2", "d2q_d2"),
            "Torsion2",
            use_mpmath=use_mpmath,
        )
        return func_result_d

    def robust_dihedral1():
        # First component of robust dihedral
        # See Eq. (3.1) in [3]
        U = M.position_wrt(O)
        V = N.position_wrt(P)
        U_ = U.normalize()
        V_ = V.normalize()
        q_rd1 = U_.dot(V_)
        dx_rd1 = (m0, m1, m2, o0, o1, o2, p0, p1, p2, n0, n1, n2)
        args_rd1 = "m0, m1, m2, o0, o1, o2, p0, p1, p2, n0, n1, n2"
        func_result_rd1 = make_deriv_funcs(
            q_rd1,
            dx_rd1,
            args_rd1,
            ("q_rd1", "dq_rd1", "d2q_rd1"),
            "RobustTorsion1",
            use_mpmath=use_mpmath,
        )
        return func_result_rd1

    def robust_dihedral2():
        # Second component of robust dihedral
        # See Eq. (3.2) in [3]
        U = M.position_wrt(O)
        W = P.position_wrt(O)
        V = N.position_wrt(P)
        U_ = U.normalize()
        W_ = W.normalize()
        V_ = V.normalize()
        q_rd2 = W_.dot(U_.cross(V_))
        dx_rd2 = (m0, m1, m2, o0, o1, o2, p0, p1, p2, n0, n1, n2)
        args_rd2 = "m0, m1, m2, o0, o1, o2, p0, p1, p2, n0, n1, n2"
        func_result_rd2 = make_deriv_funcs(
            q_rd2,
            dx_rd2,
            args_rd2,
            ("q_rd2", "dq_rd2", "d2q_rd2"),
            "RobustTorsion2",
            use_mpmath=use_mpmath,
        )
        return func_result_rd2

    def linear_bend():
        # Linear Bend
        U = M.position_wrt(O)
        V = N.position_wrt(O)
        W = P.position_wrt(Sys)
        q_lb = W.dot(U.cross(V)) / (U.magnitude() * V.magnitude())
        dx_lb = (m0, m1, m2, o0, o1, o2, n0, n1, n2)
        # Additional args, as we also supply an orthogonal direction
        args_lb = "m0, m1, m2, o0, o1, o2, n0, n1, n2, p0, p1, p2"
        func_result_lb = make_deriv_funcs(
            q_lb,
            dx_lb,
            args_lb,
            ("q_lb", "dq_lb", "d2q_lb"),
            "Linear Bend",
            use_mpmath=use_mpmath,
        )
        return func_result_lb

    def out_of_plane():
        U = M.position_wrt(P)
        V = N.position_wrt(P)
        W = O.position_wrt(P)

        U_ = U.normalize()
        W_ = W.normalize()
        V_ = V.normalize()

        Z = U_.cross(V_) + V_.cross(W_) + W_.cross(U_)
        Z_ = Z.normalize()

        q_oop = Z_.dot(U_)

        dx_oop = (m0, m1, m2, n0, n1, n2, o0, o1, o2, p0, p1, p2)
        args_oop = "m0, m1, m2, n0, n1, n2, o0, o1, o2, p0, p1, p2"
        func_result_oop = make_deriv_funcs(
            q_oop,
            dx_oop,
            args_oop,
            ("q_oop", "dq_oop", "d2q_oop"),
            "OutOfPlane",
            use_mpmath=use_mpmath,
        )
        return func_result_oop

    def linear_displacement():
        U = M.position_wrt(O)
        V = N.position_wrt(O)
        W = N.position_wrt(M)
        U_ = U.normalize()
        V_ = V.normalize()
        W_ = W.normalize()

        # Vector for cross product. For the complement X should correspond
        # to the first orthogonal direction (X = W_.cross(X)).
        X = P.position_wrt(Sys)
        # Orthogonal direction
        Y = W_.cross(X)
        Y_ = Y.normalize()

        q_ld = Y_.dot(U_) + Y_.dot(V_)
        dx_ld = (m0, m1, m2, o0, o1, o2, n0, n1, n2)
        # Additional args, as we also supply an orthogonal direction
        args_ld = "m0, m1, m2, o0, o1, o2, n0, n1, n2, p0, p1, p2"
        func_result_ld = make_deriv_funcs(
            q_ld,
            dx_ld,
            args_ld,
            ("q_ld", "dq_ld", "d2q_ld"),
            "Linear Displacement",
            use_mpmath=use_mpmath,
        )
        return func_result_ld

    if generate is None:
        generate = (
            "bond",
            "bend",
            "bend2",
            "dihedral",
            "dihedral2",
            "robust_dihedral1",
            "robust_dihedral2",
            "linear_bend",
            "out_of_plane",
            "linear_displacement",
        )

    avail_funcs = {
        "bond": bond,
        "bend": bend,
        "bend2": bend2,
        "dihedral": dihedral,
        "dihedral2": dihedral2,
        "robust_dihedral1": robust_dihedral1,
        "robust_dihedral2": robust_dihedral2,
        "linear_bend": linear_bend,
        "out_of_plane": out_of_plane,
        "linear_displacement": linear_displacement,
    }
    funcs = [avail_funcs[key] for key in generate]
    func_results = list()
    for name, func in zip(generate, funcs):
        print(f"Generating expressions for: '{name}'")
        func_res = func()
        func_results.append(func_res)

    import_str = "import mpmath" if use_mpmath else "import math"
    if out_fn:
        with open(out_fn, "w") as handle:
            handle.write(f"{import_str}\n\nimport numpy as np\n\n\n")
            for fr in func_results:
                handle.write(fr.f0 + "\n\n\n")
                handle.write(fr.f1 + "\n\n\n")
                handle.write(fr.f2 + "\n\n\n")
    print(f"Wrote generated code to '{out_fn}'")

    return func_results


if __name__ == "__main__":
    generate_wilson(out_fn="derivatives.py", use_mpmath=False)
    # print()
    generate_wilson(out_fn="mp_derivatives.py", use_mpmath=True)
    # generate_wilson(out_fn="lindisp.py", use_mpmath=False)



import numpy as np
import qcengine as qcng
import qcelemental as qcel

from pysisyphus.calculators.Calculator import Calculator


class QCEngine(Calculator):

    def __init__(self, program, model, keywords=None, connectivity=None,
                 bond_order=1, **kwargs):
        super().__init__(**kwargs)

        self.program = program
        self.model = model
        if keywords is None:
            keywords = dict()
        self.keywords = dict(keywords)
        self.connectivity = connectivity
        self.bond_order = int(bond_order)

        # TODO: pal, memory

    def get_molecule(self, atoms, coords):
        c3d = coords.reshape(-1, 3)# * BOHR2ANG
        mol_inp = {
            "symbols": atoms,
            "geometry": c3d,
            "molecular_charge": self.charge,
            "molecular_multiplicity": self.mult,
        }

        if self.program == "openmm":
            if self.connectivity is None:
                self.log( "No connectivity specified! Using hacky connectivity "
                         f"guess with bond-order={self.bond_order}")
                connectivity = qcel.molutil.guess_connectivity(atoms, c3d,
                                                               threshold=1.3)
                self.connectivity = [(at1, at2, self.bond_order)
                                     for at1, at2 in connectivity]
            mol_inp["connectivity"] = self.connectivity

        molecule = qcel.models.Molecule.from_data(mol_inp)

        return molecule

    def compute(self, molecule, driver):
        inp = {
            "molecule": molecule,
            "driver": driver,
            "model": self.model,
            "keywords": self.keywords,
        }

        result = qcng.compute(inp, program=self.program, return_dict=True)
        self.calc_counter += 1
        return result

    def keep_stdout(self, res):
        if "stdout" in res:
            with open(self.make_fn(f"qce_{self.program}_stdout"), "w") as handle:
                    handle.write(res["stdout"])

    def get_energy(self, atoms, coords, **prepare_kwargs):
        mol = self.get_molecule(atoms, coords)

        res = self.compute(mol, driver="energy")
        self.keep_stdout(res)

        results = {
            "energy": res["return_result"],
        }

        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        mol = self.get_molecule(atoms, coords)

        res = self.compute(mol, driver="gradient")
        self.keep_stdout(res)

        results = {
            "energy": res["properties"]["return_energy"],
            "forces": -np.array(res["return_result"]),
        }

        return results

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        mol = self.get_molecule(atoms, coords)

        res = self.compute(mol, driver="hessian")
        self.keep_stdout(res)

        size = 3 * len(atoms)
        shape = (size, size)

        results = {
            "energy": res["properties"]["return_energy"],
            "hessian": np.array(res["return_result"]).reshape(shape),
        }

        return results

    def __str__(self):
        return f"QCECalculator({self.name})"



"""Shared Click group helpers for lazy subcommand loading and bool normalization."""

from __future__ import annotations

import importlib
from collections.abc import Callable, Mapping

import click

from .bool_compat import normalize_argv_option_names as _normalize_argv_option_names

LazySubcommands = Mapping[str, tuple[str, str, str]]
BoolOptionsByCommand = Mapping[str, frozenset[str]]
BoolNegativeAliasesByCommand = Mapping[str, Mapping[str, str]]
BoolSingleFlagOptionsByCommand = Mapping[str, frozenset[str]]
PrimaryHelpOptionsByCommand = Mapping[str, frozenset[str]]
ParserWrapperBoolOptionProviders = Mapping[str, Callable[[], frozenset[str]]]

NormalizeBoolArgvFunc = Callable[
    [
        list[str],
        BoolOptionsByCommand,
        BoolOptionsByCommand,
        BoolNegativeAliasesByCommand,
        BoolSingleFlagOptionsByCommand,
    ],
    tuple[list[str], bool],
]
EnsureHelpAdvancedOptionFunc = Callable[[click.Command], click.Command]
ConfigureSubcommandHelpVisibilityFunc = Callable[
    [str, click.Command, PrimaryHelpOptionsByCommand], click.Command
]
BuildUnavailableCommandFunc = Callable[[str, ImportError], click.Command]


def build_unavailable_command(command_name: str, exc: ImportError) -> click.Command:
    """Return a placeholder command that reports import failure details at runtime."""
    missing = exc.name if isinstance(exc, ModuleNotFoundError) else None
    msg_lines = [
        f"Command '{command_name}' is unavailable because the module could not be imported."
    ]
    if missing:
        msg_lines.append(f"Missing dependency: {missing}")
        msg_lines.append("Install the missing dependency in your runtime environment and retry.")
    else:
        msg_lines.append(f"Import error: {exc}")

    help_text = (
        f"[Unavailable] {command_name} command.\n"
        "The command failed to import due to a missing dependency."
    )

    @click.command(name=command_name, help=help_text)
    def _unavailable() -> None:
        raise click.ClickException("\n".join(msg_lines))

    return _unavailable


class DefaultGroup(click.Group):
    """Click group with default subcommand + lazy loading + bool compatibility normalization."""

    def __init__(
        self,
        *args,
        default: str | None = None,
        lazy_subcommands: LazySubcommands | None = None,
        command_bool_value_options: BoolOptionsByCommand | None = None,
        command_bool_toggle_options: BoolOptionsByCommand | None = None,
        command_bool_toggle_negative_aliases: BoolNegativeAliasesByCommand | None = None,
        command_bool_single_flag_options: BoolSingleFlagOptionsByCommand | None = None,
        parser_wrapper_subcommands: frozenset[str] | None = None,
        parser_wrapper_bool_option_providers: ParserWrapperBoolOptionProviders | None = None,
        subcommand_primary_help_options: PrimaryHelpOptionsByCommand | None = None,
        normalize_bool_argv: NormalizeBoolArgvFunc | None = None,
        ensure_help_advanced_option: EnsureHelpAdvancedOptionFunc | None = None,
        configure_subcommand_help_visibility: ConfigureSubcommandHelpVisibilityFunc | None = None,
        build_unavailable_command: BuildUnavailableCommandFunc = build_unavailable_command,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if normalize_bool_argv is None:
            raise ValueError("normalize_bool_argv is required")
        if ensure_help_advanced_option is None:
            raise ValueError("ensure_help_advanced_option is required")
        if configure_subcommand_help_visibility is None:
            raise ValueError("configure_subcommand_help_visibility is required")

        self._default_cmd = default
        self._lazy_subcommands = dict(lazy_subcommands or {})
        self._lazy_cache: dict[str, click.Command] = {}
        self._command_bool_value_options = dict(command_bool_value_options or {})
        self._command_bool_toggle_options = dict(command_bool_toggle_options or {})
        self._command_bool_toggle_negative_aliases = dict(
            command_bool_toggle_negative_aliases or {}
        )
        self._command_bool_single_flag_options = dict(command_bool_single_flag_options or {})
        self._resolved_bool_options_by_command: dict[
            str, tuple[frozenset[str], frozenset[str], dict[str, str], frozenset[str]]
        ] = {}
        self._parser_wrapper_subcommands = set(parser_wrapper_subcommands or set())
        self._parser_wrapper_bool_option_providers = dict(
            parser_wrapper_bool_option_providers or {}
        )
        self._resolved_parser_wrapper_bool_options: dict[str, frozenset[str]] = {}
        self._subcommand_primary_help_options = dict(subcommand_primary_help_options or {})
        self._normalize_bool_argv = normalize_bool_argv
        self._ensure_help_advanced_option = ensure_help_advanced_option
        self._configure_subcommand_help_visibility = configure_subcommand_help_visibility
        self._build_unavailable_command = build_unavailable_command

    def _resolve_parser_wrapper_bool_options(self, command_name: str) -> frozenset[str]:
        cached = self._resolved_parser_wrapper_bool_options.get(command_name)
        if cached is not None:
            return cached

        provider = self._parser_wrapper_bool_option_providers.get(command_name)
        if provider is None:
            resolved = frozenset()
            self._resolved_parser_wrapper_bool_options[command_name] = resolved
            return resolved

        try:
            provided = provider()
        except Exception:
            resolved = frozenset()
            self._resolved_parser_wrapper_bool_options[command_name] = resolved
            return resolved

        normalized: set[str] = set()
        for name in provided:
            if not name.startswith("--"):
                continue
            if name.startswith("--no-"):
                normalized.add(f"--{name[5:]}")
            else:
                normalized.add(name)

        resolved = frozenset(normalized)
        self._resolved_parser_wrapper_bool_options[command_name] = resolved
        return resolved

    @staticmethod
    def _long_option_names(option_names: list[str]) -> tuple[str, ...]:
        return tuple(name for name in option_names if name.startswith("--"))

    def _resolve_bool_options(
        self, ctx: click.Context, command_name: str
    ) -> tuple[frozenset[str], frozenset[str], dict[str, str], frozenset[str]]:
        cached = self._resolved_bool_options_by_command.get(command_name)
        if cached is not None:
            return cached

        bool_value_options = set(
            self._command_bool_value_options.get(command_name, frozenset())
        )
        bool_toggle_options = set(
            self._command_bool_toggle_options.get(command_name, frozenset())
        )
        bool_toggle_negative_aliases = dict(
            self._command_bool_toggle_negative_aliases.get(command_name, {})
        )
        bool_single_flag_options = set(
            self._command_bool_single_flag_options.get(command_name, frozenset())
        )

        bool_toggle_options.update(
            self._resolve_parser_wrapper_bool_options(command_name)
        )

        command = self.get_command(ctx, command_name)
        if command is not None:
            for param in command.params:
                if not isinstance(param, click.Option):
                    continue

                positive_long_names = self._long_option_names(param.opts)
                if not positive_long_names:
                    continue

                negative_long_names = self._long_option_names(param.secondary_opts)
                if param.is_bool_flag:
                    if negative_long_names:
                        bool_toggle_options.update(positive_long_names)
                        first_negative = negative_long_names[0]
                        for index, positive_name in enumerate(positive_long_names):
                            negative_name = (
                                negative_long_names[index]
                                if index < len(negative_long_names)
                                else first_negative
                            )
                            bool_toggle_negative_aliases.setdefault(
                                positive_name, negative_name
                            )
                    else:
                        bool_single_flag_options.update(positive_long_names)
                    continue

                if isinstance(param.type, click.types.BoolParamType):
                    bool_value_options.update(positive_long_names)

        resolved = (
            frozenset(bool_value_options),
            frozenset(bool_toggle_options),
            bool_toggle_negative_aliases,
            frozenset(bool_single_flag_options),
        )
        self._resolved_bool_options_by_command[command_name] = resolved
        return resolved

    def parse_args(self, ctx, args):
        # Normalize long option names to lowercase before any other processing
        # so that --add-linkH, --include-H2O etc. are accepted case-insensitively.
        args = _normalize_argv_option_names(args)

        show_help_or_version = any(a in ("-h", "--help", "--version") for a in args)

        if self._default_cmd is not None and not show_help_or_version:
            if not args or args[0].startswith("-"):
                args.insert(0, self._default_cmd)

        bool_value_options = self._command_bool_value_options
        bool_toggle_options = self._command_bool_toggle_options
        bool_toggle_negative_aliases = self._command_bool_toggle_negative_aliases
        bool_single_flag_options = self._command_bool_single_flag_options
        if args and not args[0].startswith("-"):
            command_name = args[0]
            (
                command_bool_value_options,
                command_bool_toggle_options,
                command_toggle_negative_aliases,
                command_bool_single_flag_options,
            ) = self._resolve_bool_options(ctx, command_name)
            bool_value_options = {command_name: command_bool_value_options}
            bool_toggle_options = {command_name: command_bool_toggle_options}
            bool_toggle_negative_aliases = {
                command_name: command_toggle_negative_aliases
            }
            bool_single_flag_options = {command_name: command_bool_single_flag_options}

        args, _ = self._normalize_bool_argv(
            args,
            bool_value_options,
            bool_toggle_options,
            bool_toggle_negative_aliases,
            bool_single_flag_options,
        )
        return super().parse_args(ctx, args)

    @staticmethod
    def _silence_pysisyphus_loggers():
        """Remove pysisyphus file handlers and suppress log output.

        pysisyphus creates FileHandler + StreamHandler in its various
        ``__init__.py`` files at import time, overriding any prior
        level settings.  This method must run *after* the lazy import
        has loaded the subcommand module (and thus pysisyphus).
        """
        import logging as _logging

        _PYSIS_LOGGERS = (
            "pysisyphus", "calculator", "cos", "dimer", "dynamics",
            "gdiis", "internal_coords", "irc", "optimizer",
            "tsoptimizer", "mwfn", "stocastic", "wfoverlap",
        )
        for name in _PYSIS_LOGGERS:
            lg = _logging.getLogger(name)
            lg.setLevel(_logging.CRITICAL + 1)
            lg.propagate = False
            for h in lg.handlers[:]:
                lg.removeHandler(h)
                try:
                    h.close()
                except Exception:
                    pass

    def invoke(self, ctx):
        # Add a leading blank line for subcommands (except "all") to separate CLI tool logs.
        if ctx.invoked_subcommand and ctx.invoked_subcommand != "all":
            click.echo()
        # Suppress pysisyphus loggers AFTER lazy import has loaded the
        # subcommand module (which triggers pysisyphus __init__ file handlers).
        self._silence_pysisyphus_loggers()
        return super().invoke(ctx)

    def list_commands(self, ctx):
        cmds = set(super().list_commands(ctx))
        cmds.update(self._lazy_subcommands.keys())
        return sorted(cmds)

    def get_command(self, ctx, cmd_name):
        cmd = super().get_command(ctx, cmd_name)
        if cmd is not None:
            if cmd_name not in self._parser_wrapper_subcommands:
                cmd = self._ensure_help_advanced_option(cmd)
            return self._configure_subcommand_help_visibility(
                cmd_name, cmd, self._subcommand_primary_help_options
            )

        lazy_spec = self._lazy_subcommands.get(cmd_name)
        if lazy_spec is None:
            return None

        cached = self._lazy_cache.get(cmd_name)
        if cached is not None:
            return cached

        module_name, attr_name, _ = lazy_spec
        try:
            module = importlib.import_module(module_name, package=__package__)
            loaded_cmd = getattr(module, attr_name)
        except (ModuleNotFoundError, ImportError) as exc:
            loaded_cmd = self._build_unavailable_command(cmd_name, exc)

        if cmd_name not in self._parser_wrapper_subcommands:
            loaded_cmd = self._ensure_help_advanced_option(loaded_cmd)
        loaded_cmd = self._configure_subcommand_help_visibility(
            cmd_name, loaded_cmd, self._subcommand_primary_help_options
        )
        self._lazy_cache[cmd_name] = loaded_cmd
        return loaded_cmd

    def format_commands(self, ctx, formatter):
        rows = []
        for subcommand in self.list_commands(ctx):
            lazy_spec = self._lazy_subcommands.get(subcommand)
            if lazy_spec is not None:
                rows.append((subcommand, lazy_spec[2]))
                continue

            cmd = super().get_command(ctx, subcommand)
            if cmd is None or cmd.hidden:
                continue
            rows.append(
                (
                    subcommand,
                    cmd.get_short_help_str(formatter.width - 6 - len(subcommand)),
                )
            )

        if rows:
            with formatter.section("Commands"):
                formatter.write_dl(rows)



import numpy as np

from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.intcoords.derivatives import d2q_b
from pysisyphus.linalg import norm3


class Stretch(Primitive):

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        return Stretch.rho(atoms, coords3d, indices)

    @staticmethod
    def _calculate(coords3d, indices, gradient=False):
        n, m = indices
        bond = coords3d[m] - coords3d[n]
        bond_length = norm3(bond)
        if gradient:
            bond_normed = bond / bond_length
            row = np.zeros_like(coords3d)
            # 1 / -1 correspond to the sign factor [1] Eq. 18
            row[m,:] =  bond_normed
            row[n,:] = -bond_normed
            row = row.flatten()
            return bond_length, row
        return bond_length

    @staticmethod
    def _jacobian(coords3d, indices):
        return d2q_b(*coords3d[indices].flatten())
